#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| @purpose: Hub for creating plots.  
| @date: Created on Sat May 1 15:12:38 2019   
| @author: Semeon Risom   
| @email: semeon.risom@gmail.com   
| @url: https://semeon.io/d/mdl
"""

# available functions
__all__ = ['bokeh_trial','bokeh_calibration','onset_diff_plot','density_plot','corr_matrix','boxplot','cooks_plot',
'residual_plot','qq_plot','logit_plot','html']

# core
from pdb import set_trace as breakpoint
from distutils import dir_util
import importlib
import datetime
import shutil
import sys
import os

# data
import pandas as pd
import numpy as np

# seaborn
import matplotlib
matplotlib.use('Agg')
import seaborn as sns
import matplotlib.pyplot as plt

# local libraries
from . import settings

# constants
console = settings.console
debug = settings.debug

def __font__():
	"""Add Helvetica to matplotlib."""
	from matplotlib import matplotlib_fname, rcParams
	import matplotlib.font_manager as font_manager

	directory = matplotlib_fname().replace("/matplotlibrc/", "")
	destination = f'{directory}/fonts/ttf'
	file = settings.path['home'] + "/dist/resources/Helvetica.ttf"

	#add to matplotlib font folder
	shutil.copy(file, destination)

	#add to computer font folder
	##if running osx
	if sys.platform == "darwin":
		shutil.copy(file, '/Library/Fonts/')
	##if running win32
	if sys.platform == "win32":
		shutil.copy(file, 'c:\\windows\\fonts')

	#rebuild fonts
	prop = font_manager.FontProperties(fname=file)
	prop.set_weight = 'light'
	rcParams['font.family'] = prop.get_name()
	rcParams['font.weight'] = 'light'
	font_manager._rebuild()

def bokeh_trial(config, df, stim_bounds, roi_bounds, flt):
    """Create single subject trial bokeh plots.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        Pandas dataframe of participant sample data.
    stim_bounds : :class:`dict`
        Stimulus bounds on screen.
    roi_bounds : :class:`dict`
        ROI bounds on screen.
    flt : :class:`str`
        Filter type.
    """
	#----bokeh
    from bokeh.core.properties import value
    from bokeh.plotting import reset_output, figure
    from bokeh.models import HoverTool, Range1d, BoxAnnotation, ColumnDataSource, CDSView, BooleanFilter
    from bokeh.layouts import gridplot
    from bokeh.embed import components

    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    #metadata
    emotion = [df['left_mood'][0],df['right_mood'][0]]
    #filter xy coordinates
    flt_x = '%s_x'%(flt)
    flt_y = '%s_y'%(flt)
    
    #get display size
    xy = [int(i) for i in (df['monitorSize.px'].values[0]).split('x')]
    monitorSize = xy
    
    #subject data
    t = df['timestamp'].values
    
    #get fixations
    df['fixation_onset'] = df.loc[df.groupby('fix_num')['%s_fix_all'%(flt)].head(1).index, '%s_fix_all'%(flt)]
    df['fixation_onset'] = df.apply(lambda x: True if (x['fixation_onset'] == True) else False, axis=1)
    
    #source for syncing data
    s = ColumnDataSource(df)

    ##tools = "box_select,save,pan,box_zoom,reset"
    tools="box_select,save,reset"

    #figure 1
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    p1 = figure(
            output_backend="webgl",
            tools=tools,
            plot_width=1200,
            plot_height=400,
            x_range=Range1d(-100, (t[-1]+100), bounds="auto"),
            y_range=Range1d(0,(monitorSize[0]+100), bounds="auto"),
            x_axis_label = 'time (msec)', y_axis_label = 'x/y (px)',
            sizing_mode="stretch_both",
     )

    #hover
    hover = HoverTool()
    hover.tooltips = [
        ("name", "$name"),
        ("sample", "$index"),
        ("data (x,y)", "($x{(0)},$y{(0)})")
    ]
    p1.tools.append(hover)
    p1.hover.line_policy = "nearest"

    #graph
    ###-----------------------------------------------------------------x=time, y=(x or y)
    #original x
    p1.scatter(x='timestamp', y='x', color='#2965C0', alpha=0, line_width=2,
            line_dash='dashed', name='raw_x', source=s)
    p1.line(x='timestamp', y='x', color='#2965C0', alpha=0.5, line_width=2,
            legend=value('raw_x'), line_dash='dashed', name='raw_x', source=s)
    #filtered x
    p1.scatter(x='timestamp', y=flt_x, color='#265EB3', alpha=0, line_width=2,
            name=flt_x, source=s)
    p1.line(x='timestamp', y=flt_x, color='#265EB3', alpha=1, line_width=2,
            legend=value(flt_x), name=flt_x, source=s)
    #peak removed
    #p1.line(x=t, y=sg_x, color='#101f23', alpha=1, line_width=2,legend='x_peaksg', name='x_peaksg')

    #original y
    p1.scatter(x='timestamp', y='y', color='#C01F1F', alpha=0, line_width=2,
            line_dash='dashed', name='raw_y', source=s)
    p1.line(x='timestamp', y='y', color='#C01F1F', alpha=0.5, line_width=2,
            legend=value('raw_y'), line_dash='dashed', name='raw_y', source=s)
    #filtered y
    p1.scatter(x='timestamp', y=flt_y, color='#A61B1B', alpha=0, line_width=2,
            name=flt_y, source=s)
    p1.line(x='timestamp', y=flt_y, color='#A61B1B', alpha=1, line_width=2,
            legend=value(flt_y), name=flt_y, source=s)

    #task-events (stimulus onset, dotloc onset)
    ##stimulus onset
    time = df.loc[df['marker']=='Stimulus Onset']['timestamp'].iloc[0]
    box = BoxAnnotation(left=time, right=time + 10, fill_alpha=1, fill_color='#434e54', level='glyph')
    p1.add_layout(box)
    ##dotloc onset
    time = df.loc[df['marker']=='Dotloc Onset']['timestamp'].iloc[0]
    box = BoxAnnotation(left=time, right=time + 10, fill_alpha=1, fill_color='#434e54', level='glyph')
    p1.add_layout(box)

    #fixations
    for index, row in df[~df['sg_fix_all'].isin([False])].groupby('fix_num'):
        start = row.iloc[0]['timestamp']
        end = row.iloc[-1]['timestamp']
        box = BoxAnnotation(left=start, right=end,fill_alpha=0.2, fill_color='#9e9e9e', level='glyph')
        p1.add_layout(box)
    del index, row

    #legend
    p1.toolbar.logo = None
    p1.legend.orientation = "horizontal"
    p1.legend.location = "top_left"
    p1.legend.click_policy = "hide"

    ###-----------------------------------------------------------------window display
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    #display what participant sees
    tools = "box_select,save,pan,box_zoom,reset"
    p2 = figure(
            output_backend="webgl",
            tools=tools,
            plot_width=1200,
            plot_height=600,
            x_axis_label = 'x (px)', y_axis_label = 'y (px)',
            x_range=Range1d(0, monitorSize[0], bounds="auto"),
            y_range=Range1d(0, monitorSize[1], bounds="auto"),
            sizing_mode="stretch_both"
    )
    #hover
    hover = HoverTool()
    hover.tooltips = [
        ("name", "$name"),
        ("sample", "$index"),
        ("data (x,y)", "($x{(0)},$y{(0)})")
    ]
    p2.tools.append(hover)
    p2.hover.line_policy = "nearest"
    ###---------line plots
    #------gaze xy
    p2.cross(x=flt_x, y=flt_y, size=10, color="firebrick", alpha=1, fill_alpha=0.6,
              fill_color="firebrick", source=s, name='gaze', legend='gaze')

    #------fixation onset within and outside roi
    #get booleanfilter
    l_fix_all = [True if x !=False else False for x in s.data['fixation_onset']]
    fix_all = BooleanFilter(l_fix_all)
    view = CDSView(source=s, filters=[fix_all])
    ##draw square ##2222b2
    p2.square(x=flt_x, y=flt_y, size=10, color="#2222b2", alpha=1, fill_alpha=0.6,
              fill_color="white", source=s, view=view, name='fixation', legend=value("fixation"))
    
    #------fixation within and outside roi
    #get booleanfilter
    l_fix_false = [True if x !=False else False for x in s.data['sg_fix_all']]
    fix_false = BooleanFilter(l_fix_false)
    view = CDSView(source=s, filters=[fix_false])
    ##draw square ##2222b2
    p2.square(x=flt_x, y=flt_y, size=10, color="black", alpha=1, fill_alpha=0.6,
              fill_color="white", source=s, view=view, name='false-fixation', legend=value("false-fixation"))

    #------only fixation within roi
    #get booleanfilter
    l_fix_roi = [True if (y==True and z==True) else False for y,z in zip(s.data['fixation_onset'], s.data['dwell'])]
    fix_roi = BooleanFilter(l_fix_roi)
    view = CDSView(source=s, filters=[fix_roi])
    ##draw triangle ##198D40
    p2.triangle(x=flt_x, y=flt_y, size=10, color="#198D40", alpha=1, fill_alpha=0.6,
              fill_color="white", source=s, view=view, name='roi', legend=value("roi"))
    
    #------dwell-time gaze within roi
    l_dwell_roi = [True if x !=False else False for x in s.data['dwell']]
    dwell_roi = BooleanFilter(l_dwell_roi)
    view = CDSView(source=s, filters=[dwell_roi])
    ##draw circle ##ff9800
    p2.circle(x=flt_x, y=flt_y, size=10, color="#ff9800", alpha=1, fill_alpha=0.6,
              fill_color="white", source=s, view=view, name='dwell', legend=value("dwell"))

    #------stim bounds
    stimBounds = stim_bounds

    #------roi bounds
    roiBounds = roi_bounds

    #emotion
    left, right = emotion
    if left == 'Sad':
        left_color = '#ff0000'
        right_color = '#9e9e9e'
    else:
        right_color = '#ff0000'
        left_color = '#9e9e9e'

    #roi
    #left
    p2.add_layout(BoxAnnotation(left=roiBounds[0]['bx1'], right=roiBounds[0]['bx2'],
                                top=roiBounds[0]['by1'], bottom=roiBounds[0]['by2'],
                        fill_alpha=0.05, fill_color='blue', level='glyph'))
    #right
    p2.add_layout(BoxAnnotation(left=roiBounds[1]['bx1'], right=roiBounds[1]['bx2'],
                                top=roiBounds[1]['by1'], bottom=roiBounds[1]['by2'],
                        fill_alpha=0.05, fill_color='blue', level='glyph'))


    #--------stim bounds
    #left
    p2.add_layout(BoxAnnotation(left=stimBounds[0]['bx1'], right=stimBounds[0]['bx2'],
                                top=stimBounds[0]['by1'], bottom=stimBounds[0]['by2'],
                        fill_alpha=0.3, fill_color=left_color, level='glyph'))
    #right
    p2.add_layout(BoxAnnotation(left=stimBounds[1]['bx1'], right=stimBounds[1]['bx2'],
                                top=stimBounds[1]['by1'], bottom=stimBounds[1]['by2'],
                        fill_alpha=0.3, fill_color=right_color, level='glyph'))
    #hover
    hover = HoverTool()
    hover.tooltips = [
        ("time", "@timestamp"),
        ("sample", "$index"),
        ("data (x,y)", "($%s_x{(0)},$%s_y{(0)})"%(flt,flt))
    ]
    #p2.tools.append(hover)
    #p2.hover.line_policy = "nearest"

    #legend
    p2.toolbar.logo = None
    p2.legend.orientation = "horizontal"
    p2.legend.location = "top_left"
    p2.legend.click_policy = "hide"

    #reverse axis
    p2.x_range = Range1d(0, monitorSize[0])
    p2.y_range = Range1d(monitorSize[1], 0)

    ###-----------------------------------------------------------------slider
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    start = df.iloc[0]['timestamp']
    end = df.iloc[-1]['timestamp']
    # callback = CustomJS(args=dict(source=s), code="""
    #     data = source.data;
    #     t = cb_obj;
    #     f = cb_obj.value;
    #     x = data['x'];
    #     y = data['y'];
    #     source.change.emit();
    #     console.log('test');
    # """)
    # p3 = Slider(
    #         title="time",
    #         start=start,
    #         end=end,
    #         value=1, step=1,
    #         callback=callback,
    #         bar_color='red',
    #         width=1200,
    #         show_value=True,
    #         sizing_mode="stretch_both"
    # )

    ###-----------------------------------------------------------------combining plots
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    # grid = gridplot(children=[[p1], [p2], [p3]],
    grid = gridplot(children=[[p1], [p2]],
                    merge_tools=True,
                    toolbar_location='right',
                    sizing_mode="fixed")

    #get html
    script, div = components(grid)

    ##convert seperate plots and div to single string
    plots = (''.join(map(str, [div, '\n', script])))

    #clear cache
    reset_output()

    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plots

def bokeh_calibration(config, df, cxy, event, monitorSize=[1920,1080]):
    """Create calibration matrix, using pandas and bokeh.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    cxy : :class:`pandas.DataFrame`
        Pandas dataframe of calibration points.
    event : :obj:`string`
        calibration, or validation.
    monitorSize : :obj:`list`
        Monitor size, in pixels.
    """
    from bokeh.plotting import reset_output, figure
    from bokeh.models import Range1d, ColumnDataSource
    from bokeh.embed import components

    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    #subject data
    event_t = df['event_trial'].values
    
    #source for syncing data
    s = ColumnDataSource(df)

    ##tools = "box_select,save,pan,box_zoom,reset"
    #figure 1
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    ##tools = "box_select,save,pan,box_zoom,reset"
    tools = "box_select,reset"
    p1 = figure(
            output_backend="webgl",
            tools=tools,
            plot_width=1200,
            plot_height=150,
            x_range=Range1d(0 - 0.1, (event_t[-1] + 0.1), bounds="auto"),
            y_range=Range1d(0,(monitorSize[0]+100), bounds="auto"),
            x_axis_label = 'calibration point', y_axis_label = 'x/y (px)',
            sizing_mode="scale_width",
     )

    #graph
    ###-----------------------------------------------------------------x=time, y=(x or y)
    #original x
    p1.scatter(x='event_trial', y='cx', color='#2965C0', alpha=0.5, line_width=2,
            line_dash='dashed', name='cx', source=s)
    # p1.line(x='event_trial', y='cx', color='#2965C0', alpha=0.5, line_width=2,
            # line_dash='dashed', name='cx', source=s,legend=value('cx'))

    #original y
    p1.scatter(x='event_trial', y='cy', color='#C01F1F', alpha=0.5, line_width=2,
            line_dash='dashed', name='cy', source=s)
    # p1.line(x='event_trial', y='cy', color='#C01F1F', alpha=0.5, line_width=2,
    #         line_dash='dashed', name='cy', source=s,legend=value('cy'))

    ###-----------------------------------------------------------------window display
    #sizing_mode: ``"fixed"``, ``"stretch_both"``, ``"scale_width"``, ``"scale_height"``, ``"scale_both"``
    ##tools = "box_select,save,pan,box_zoom,reset"
    tools="save,reset"
    p2 = figure(
        output_backend="webgl",
        tools=tools,
        plot_width=1200,
        plot_height=600,
        x_axis_label = 'x (px)', y_axis_label = 'y (px)',
        x_range=Range1d(0, monitorSize[0], bounds="auto"),
        y_range=Range1d(0, monitorSize[1], bounds="auto"),
        sizing_mode="scale_width"
    )
    
    ###---------line plots
    #----gaze xy
    # cross = p2.cross(x='gx', y='gy', size=10, color="firebrick", alpha=1, fill_alpha=0.6,
    p2.cross(x='gx', y='gy', size=10, color="firebrick", alpha=1, fill_alpha=0.6,
              fill_color="firebrick", source=s, name='gaze')
    #           fill_color="firebrick", source=s, name='gaze', legend=value("gaze"))
    
    #----circle xy
    p2.circle(x=cxy['cx'], y=cxy['cy'], size=12, color="#198D40", alpha=1, fill_alpha=0.6,
              fill_color="white", name='circle')
              #fill_color="white", source=s, name='circle', legend=value("circle"))
              
    # draw mouse click for calibration
    if event=='calibration':
        #----mouse xy
        p2.square(x='mx', y='my', size=10, color="#2222b2", alpha=1, fill_alpha=0.6,
                  fill_color="white", source=s, name='mouse')
                  #fill_color="white", source=s, name='mouse', legend=value("mouse"))
        
    #----legend
    p1.toolbar.logo = None
    p2.toolbar.logo = None

    #----reverse axis
    p2.x_range = Range1d(0, monitorSize[0])
    p2.y_range = Range1d(monitorSize[1], 0)
    
    #----grid
    #p1
    p1.grid.grid_line_alpha = 0.5
    p1.grid.grid_line_color = "white"
    
    #----font
    # p2
    ## title
    p2.axis.axis_label_text_font_style = "normal"
    ## labels
    p2.axis.axis_label_text_font_size= "11.5pt"
    p2.axis.axis_label_text_font = "Helvetica"
    p2.axis.axis_label_text_color = "#1c1d1e"
    p2.axis.major_label_text_font_size= "11.5pt"
    p2.axis.major_label_text_font = "Helvetica"
    p2.axis.major_label_text_color = "#1c1d1e"
    # p1
    ## title
    p1.axis.axis_label_text_font_style = "normal"
    ## labels
    p1.axis.axis_label_text_font_size = "9.5pt"
    p1.axis.axis_label_text_font = "Helvetica"
    p1.axis.axis_label_text_color = "#1c1d1e"
    p1.axis.major_label_text_font_size = "9.5pt"
    p1.axis.major_label_text_font = "Helvetica"
    p1.axis.major_label_text_color = "#1c1d1e"
    p1.axis.major_tick_line_color = None
    p1.axis.minor_tick_line_color = None 
    
    #----padding
    p2.min_border_top = 16
    p1.min_border_bottom = 0
    
    #----axis line
    # p2
    p2.axis.axis_line_width = 0
    p2.axis.axis_line_color = None
    p2.outline_line_color = "#1c1d1e"
    # p1
    p1.background_fill_color = "#dfe1e4"
    p1.axis.axis_line_width = 0
    p1.axis.axis_line_color = None
    p1.outline_line_color = "#1c1d1e"
    
    #----get html
    #get componants
    plot = {'top': p2, 'bottom': p1}
    script, div = components(plot)
    #combine
    div = ''.join(list(div.values()))
    plot = div + script
        
    #----clear cache
    reset_output()

    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plot

def onset_diff_plot(config, df, meta, drop, y, clip=None):
    """Plot onset differences using pandas and seaborn.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    meta : :class:`pandas.DataFrame`
        Metadata for each chart.
    drop : :class:`pandas.DataFrame`
        Participants to be dropped.
    y : :obj:`str`
        Variable of interest.
    clip : :obj:`int`
        Clip value for single subject plot.

    Returns
    -------
    odp
        Bokeh or seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #setting clip
    clip = 200 if clip == None else clip
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    importlib.reload(plt); importlib.reload(sns)
    console('running onset_diff_plot(%s)'%(y), 'blue')
    html_plots = []
    
    #-----------------participant trend line
    individual = meta['individual']
    sns.set(style=config['style']['seaborn'], font_scale=1.25, font="Helvetica")
    #sns.despine(offset=10, trim=True)
    fig, ax = plt.subplots(figsize=(10,5))
    #plot
    df["%s (msec)"%(y)] = df.groupby('participant')[y].rolling(window=5).mean().reset_index(0,drop=True)
    ##if diff_dotloc and diff_stim < 500 msec
    sns.color_palette("RdBu",6)
    ax_ = sns.lineplot(x="TrialNum_", y="%s (msec)"%(y), ax=ax, legend=False, lw=1, estimator=None, palette="ch:2.5,.25",
                       hue="participant", units="participant", data=df.loc[df['onset_greater'] == False])
    ##if diff_dotloc and diff_stim > 500 msec
    ax_ = sns.lineplot(x="TrialNum_", y="%s (msec)"%(y), ax=ax, legend=False, lw=1, estimator=None, alpha=0.2,
                      hue="participant", units="participant", data=df.loc[df['onset_greater'] == True])
    #settings
    ax_.set(xlim=(0, 1), ylim=(0, clip), ylabel='| %s (msec) |'%(y), xlabel='trial')
      
    #append
    html_plots.append({"title":individual['title'],"file":"%s.png"%(individual['file']),"footnote":individual['footnote'],"anchor":"itl"})
    #tight layout
    fig.tight_layout()
    #save
    fig.savefig(individual['path'], dpi=300)
    #clear figure from memory
    plt.close(fig)
    
    #-----------------diff plot
    group = meta['group']
    #for each plot
    importlib.reload(plt); importlib.reload(sns)
    fig, ax = plt.subplots(figsize=(10,10), ncols=2, nrows=2, sharey=False)
    ax1, ax2, ax3, ax4 = ax.flatten()
    sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
    #sns.despine(offset=10, trim=True)
    #binned
    ##iaps
    type_='iaps'
    df_1 = df.loc[df['trialType'] == type_].reset_index(level=0, drop=True)
    sns.regplot(x="TrialNum_",y=y,color="#5874a6",data=df_1,ax=ax1,
                fit_reg=True,x_bins=group['binned']['bins'],scatter_kws={"s": 40},line_kws={'color': 'red'})
    ax1.set(ylim=(0, 200), xlabel='', ylabel='| %s (msec) |'%(y), title='trial_type=%s'%(type_))
    ##pofa
    type_='pofa'
    df_1 = df.loc[df['trialType'] == type_].reset_index(level=0, drop=True)
    sns.regplot(x="TrialNum_",y=y,color="#5d9f6c",data=df_1,ax=ax2,
                fit_reg=True,x_bins=group['binned']['bins'],scatter_kws={"s": 40},line_kws={'color': 'red'})
    ax2.set(ylim=(0, 200), xlabel='', ylabel='', title='trial_type=%s'%(type_))
    #unbinned
    ##iaps
    type_='iaps'
    df_1 = df.loc[df['trialType'] == type_].reset_index(level=0, drop=True)
    sns.regplot(x="TrialNum_",y=y,color="#5874a6",data=df_1,ax=ax3,
                fit_reg=True,x_bins=group['unbinned']['bins'],scatter_kws={"s": 5},line_kws={'color': 'red'})
    ax3.set(ylim=(0, 1000), xlabel='trial', ylabel='| %s (msec) |'%(y), title='trial_type=%s'%(type_))
    ##pofa
    type_='pofa'
    df_1 = df.loc[df['trialType'] == type_].reset_index(level=0, drop=True)
    sns.regplot(x="TrialNum_",y=y,color="#5d9f6c",data=df_1,ax=ax4,
                fit_reg=True,x_bins=group['unbinned']['bins'],scatter_kws={"s": 5},line_kws={'color': 'red'})
    ax4.set(ylim=(0, 1000), xlabel='trial', ylabel='', title='trial_type=%s'%(type_))
    #append
    html_plots.append({"title":group['title'],"file":"%s.png"%(group['file']),"footnote":group['footnote'],"anchor":"gtl"})
    #tight layout
    fig.tight_layout()
    #save
    fig.savefig(group['path'], dpi=300)
    #clear figure from memory
    plt.close(fig)
    
    #--------finished
    #timestamp
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return html_plots

def density_plot(config, df, title):
    """
	Create density plot (draws kernel density estimate), using seaborn and pandas.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    title : :obj:`str`
        Chart title.

    Returns
    -------
    cm
        Bokeh or seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    #seaborn
    #https://seaborn.pydata.org/generated/seaborn.distplot.html
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)
    console('running density_plot()', 'blue')
    
    #remove cesd_group, trialType_ column
    df = df.drop(['trialType_','cesd_group'], axis=1)
    
    #create metadata to be used for html file
    html_plots = []
    
    #for each column
    for item in (df.columns.tolist()):
        #----plot type
        sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
        sns.despine(offset=10, trim=True)

        #----plot
        fig, ax = plt.subplots(figsize=(10,10))
        density = sns.distplot(df[item], ax=ax, norm_hist=True,
                               #bins - number of bins for histogram
                               hist=True, bins=20,
                               #bw (bandwidth) - width of the kernels
                               kde=True, kde_kws={'clip':None,'bw':'scott','kernel':'gau'}).set(xlim=(None, None))
        ax.set(ylabel='Density')
        
        #----design
        #tight layout
        fig.tight_layout()
        
        #----save
        #check if path exists
        file = "density_%s.png"%(item)
        path = config['path']['output'] + "/analysis/html/img/density/"
        if not os.path.exists(path):
            os.makedirs(path)
        #save
        fig.savefig(path + file, dpi=300)
        
        #append 
        html_plots.append({"title":'%s for %s.'%(title, item),"file":'density/'+file,"footnote":""})
        #clear figure from memory
        plt.close(fig)
         
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return density, html_plots

def corr_matrix(config, df, path, title, method, footnote=None):
    """
	Create correlation matrix using bokeh and pandas.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    path : :obj:`str`
        The directory path to save the bokeh or plot.
    method : :obj:`str`
        Spearman or Pearsons correlation coefficient.
    title : :obj:`str`
        Chart title.
    footnote : :obj:`str`
        Chart footnote.

    Returns
    -------
    cm
        Bokeh plot.
    """
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('running bokeh corr_matrix()', 'blue')

    # math
    import bisect
    from math import pi
    from numpy import arange
    from itertools import chain
    from collections import OrderedDict
    
    # bokeh
    from bokeh.models import ColorBar, LinearColorMapper, TapTool, HoverTool, Range1d, ColumnDataSource
    from bokeh.models.callbacks import CustomJS
    from bokeh.plotting import reset_output, figure
    from bokeh.embed import components

    #create color palette
    matplotlib.use('Agg')
    import matplotlib.colors
    from matplotlib import cm as mpl_cmap
    
    #p-values
    if method == "pearson":
        from scipy.stats import pearsonr as cf
    elif method == "spearman":
        from scipy.stats import spearmanr as cf
    
    #calculate p-values correlation coefficent
    def get_pvalue(df):
        df = df.dropna()._get_numeric_data()
        dfcols = pd.DataFrame(columns=df.columns)
        pvalues = dfcols.transpose().join(dfcols, how='outer')
        for row in df.columns:
            for column in df.columns:
                pvalues[row][column] = round(cf(df[row], df[column])[1], 4)
        return pvalues

    #Gets bounds for quads with n features
    def get_bounds(n):
        bottom = list(chain.from_iterable(
            [[ii]*nlabels for ii in range(nlabels)]))
        top = list(chain.from_iterable(
            [[ii+1]*nlabels for ii in range(nlabels)]))
        left = list(chain.from_iterable(
            [list(range(nlabels)) for ii in range(nlabels)]))
        right = list(chain.from_iterable(
            [list(range(1, nlabels+1)) for ii in range(nlabels)]))
        corr_items = list(chain.from_iterable(
            [[ii+1]*nlabels for ii in range(nlabels)]))
        return top, bottom, left, right, corr_items
    
    #Aligns color values from palette with the correlation coefficient and p-values
    def get_colors_corr(corr_array, p_array, colors):
        c_corr = arange(-1, 1, 1/(len(colors)/2))
        corr_color = []
        p_color = []
        corr_list = []
        p_list = []
        factor_list = []
        itr = 0
        for corr, pvalue in zip(corr_array,p_array):
            ind = bisect.bisect_left(c_corr, corr)
            #colors
            ##corr_color
            if (itr)%(nlabels+1)==0:
                corr_color.append('#022f62')
            else:
                corr_color.append(colors[ind-1])
            ##pcolor
            if pvalue<=0.05:
                p_color.append("#F44336")
            else:
                p_color.append(colors[ind-1])
            #append corr and pvalues
            corr_list.append(corr)
            p_list.append(pvalue)
            factor_list.append(['[%s, %s]'%(l_labels[itr][0],l_labels[itr][1])])
            itr = itr + 1
        return corr_color, corr_list, p_color, p_list, factor_list
    
    #create seaborn plots for each possible correlation combinations
    def get_corr_plot(df, x, y):
        importlib.reload(plt); importlib.reload(sns)

        #----parameters
        #figure
        fig, ax = plt.subplots(1, 1, figsize=(10,10))
        #bounds
        plt.subplots_adjust(top=0.95, bottom=0.085, left=0.125, right=0.975, hspace=0.2, wspace=0.2)
        
        #----title
        title = "x=%s, y=%s"%(x,y)
        
        #----set style
        #pal = sns.color_palette("Set1", n_colors=2, desat=.5)
        sns.set(style=config['style']['seaborn'], font="Helvetica", font_scale=1.1)
        sns.despine(offset=10, trim=True)
        
        #----plot
        #if grouping #g = sns.lmplot(x, y, hue=groupby, data=df, palette=pal).set_title(title)
        sns.regplot(x, y, data=df, scatter=True, fit_reg=True, ci=95).set_title(title)
        
        #----add rho, p-value to plot
        #get values
        corr_r = cf(df[x],df[y])
        corr_t = "r = %.2f, p < %.2f"%(corr_r[0],corr_r[1])
        #place text 90% of max-x and 90% of max-y
        fig.text(0.85, 0.05, corr_t, size=16, horizontalalignment='center',
                verticalalignment='center', transform=ax.transAxes)

        #----save
        #check if path exists
        file = "%s.png"%(title)
        path = config['path']['output'] + "/analysis/html/img/corr/"
        if not os.path.exists(path):
            os.makedirs(path)
        #save
        fig.savefig(path + file, dpi=300)

    #reset
    reset_output()
    
    #get color map
    RdYlBu = mpl_cmap.get_cmap('RdBu', 32)
    colors_np = np.vstack(RdYlBu(np.linspace(0, 1, 32)))
    cmap = matplotlib.colors.ListedColormap(colors_np, name='RedBlue')
    colors = [matplotlib.colors.rgb2hex(cmap(x/32)[:3]) for x in range(32)]

    # calculate correlation coefficients
    #coeff
    corr_coeff = df.corr(method=method)
    #p-value
    p_value = get_pvalue(df)

    # get list and number of variables
    labels = df.columns
    nlabels = len(corr_coeff)

    #create list of labels
    ##x
    l_labels_x = labels.tolist() * nlabels
    ##y
    l_labels_y = labels.tolist() * nlabels
    ###split into chunks
    l_labels_y_split = np.hsplit(np.array(np.array_split(l_labels_y, nlabels)), nlabels)
    l_labels_y = np.concatenate(l_labels_y_split).ravel().tolist()
    ##combine
    l_labels = list(zip(l_labels_x, l_labels_y))

    #create coeff label coordinates
    num = np.arange(0, nlabels, 0.5).tolist()
    num = [x for x in num if x not in np.arange(0, nlabels).tolist()]
    ##x
    num_x = [(x - 0.025) for x in num]
    num_x = [x for x in num_x if x not in np.arange(0, nlabels).tolist()] * nlabels
    ##y
    num_y = [(x + .125) for x in num]
    num_y = [y for y in num_y if y not in np.arange(0, nlabels).tolist()] * nlabels

    ###split into chunks
    num_y_split = np.hsplit(np.array(np.array_split(num_y, nlabels)), nlabels)
    num_y = np.concatenate(num_y_split).ravel().tolist()
    ##combine
    #num_labels = list(zip(num_x, num_y))

    #plot
    tools="box_select,save,reset"
    cm = figure(tools=tools, plot_width=1200, plot_height=1200, x_range=(0, nlabels), y_range=(0, nlabels), 
                output_backend="webgl")

    #hover
    hover = HoverTool()
    hover.tooltips = [
         ("factor", "@factor"),
         ("rho", "@coeff_num"),
         ("p", "@p_num"),
    ]
    #cm.tools.append(hover)
    cm.hover.line_policy = "nearest"

    #legend
    #cm.toolbar.logo = None

    #grid
    cm.xgrid.grid_line_color = None
    cm.ygrid.grid_line_color = None
    cm.xaxis.major_label_orientation = pi/4
    cm.yaxis.major_label_orientation = pi/4

    #turn off scientific notation
    cm.left[0].formatter.use_scientific = False

    # prepare squares for plot
    top, bottom, left, right, corr_items = get_bounds(nlabels)
    color_list,corr,p_color,pvalues,factor = get_colors_corr(corr_coeff.values.flatten(),p_value.values.flatten(),colors)

    #prepare corr_coeff label for plot
    coeff_color = ["#444444"] * (nlabels * nlabels)
    coeff_color = ['#f9f9f9' if cr >= np.float64(0.74) else cof for cof, cr in zip(coeff_color, corr)]
    coeff_num = ["%.2f"%(x) for x in corr]

    #prepare pvalue for plot
    p_color = ["#444444"] * (nlabels * nlabels)
    p_color = ['#F44336' if p <= np.float64(0.05) else cof for pcol, cof, p in zip(p_color, coeff_color, pvalues)]
    p_num = ["(%.2f)"%(y) for y in pvalues]

    #data source
    source = ColumnDataSource(data=dict(left=left, right=right, top=top, bottom=bottom,
                                        x=num_x, y=num_y, coeff_num=coeff_num, coeff_color=coeff_color,
                                        p_num=p_num, factor=factor, square_color=color_list, p_color=p_color))
    #create squares
    cm.quad(top='top', bottom='bottom', left='left', right='right', 
            line_color='white', color='square_color', source=source)

    #create text
    #correlation coefficient
    cm.text(x='x', y='y', source=source, text_color="coeff_color", text='coeff_num', 
            text_font_style='normal', text_font_size = '1.5em', text_align='center',
            y_offset=-5, x_offset=1, text_line_height=1)
    #p-value
    cm.text(x='x', y='y', source=source, text_color="p_color", text='p_num', 
            text_font_style='normal', text_font_size = '1.0em', text_align='center',
            y_offset=20, x_offset=1, text_line_height=1)

    #callback
    link = CustomJS(args = dict(source = source), code="""
        console.log("second")
        obj = cb_obj
        s = source
        data = source.data
        attr = source.attributes
        d = attr.data

        //selected item
        id = source.selected['1d'].indices
        item = d.factor[id][0].replace(/[[\]]/g,'').match(/(".*?")|(\S+)/g)
        console.log('id: ' + id + ', item: ' + item)
        
        //url
        path = window.location.href
        current = path.split("/").slice(0,-1).join("/")
        url = current + '/reg.html?x='+ item[0] + 'y=' + item[1]
        window.open(url)
        //window.open(url, 'newwindow', config='height=720, width=1280')
    """)
    cm.add_tools(TapTool(callback=link))  

    #reverse y-axis
    cm.y_range = Range1d(nlabels, 0)
    #p.y_range = Range1d(monitorSize[1], 0)

    # Set ticks with labels
    ticks = [tick+0.5 for tick in list(range(nlabels))]
    tick_dict = OrderedDict([[tick, labels[ii]] for ii, tick in enumerate(ticks)])

    # Create the correct number of ticks for each axis
    cm.xaxis.ticker = ticks
    cm.yaxis.ticker = ticks

    # Override the labels
    cm.xaxis.major_label_overrides = tick_dict
    cm.yaxis.major_label_overrides = tick_dict

    #color bar
    mapper = LinearColorMapper(
        palette=colors,
        low=-1, high=1
    )
    color_bar = ColorBar(color_mapper=mapper, location=(0, 0))
    cm.add_layout(color_bar, 'right')

    #create link for each comparison
    all_columns = [[x,y] for x in list(labels) for y in list(labels)]
    for idx, clm in enumerate(all_columns):
        get_corr_plot(df=df, x=clm[0], y=clm[1])

    #get html
    script, div = components(cm)
    
    ##convert seperate plots and div to single string
    plots = (''.join(map(str, [div, '\n', script])))

    #create html
    html(config, path=path, plots=plots, source='bokeh', title=title, footnote=footnote)

    #reset
    reset_output()

    #--------finished
    #timestamp
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return cm

def single_subject(self, df, path):
    """Create single subject scatterplot using seaborn and pandas.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    path : :obj:`str`
        The directory path to save the bokeh or seaborn plot.

    Returns
    -------
    cm
        Bokeh or seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('running single_subject()', 'blue')
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)
    
    #styling plot
    sns.set(style="darkgrid", font_scale=1, font="Helvetica")
    sns.set_context("paper")

    #events
    events = self.config['events']
    
    #get first pofa and iaps
    first = df.drop_duplicates('trialType', keep='first')['TrialNum'].tolist()
    
    #get display size
    xy = [int(i) for i in (df['monitorSize.px'].values[0]).split('x')]
    
    #drop x - outside coordinates
    df = df.loc[~(df['sg_x']<0)] #drop all gaze_x < 0
    df = df.loc[~(df['sg_x']>xy[0])] #drop all gaze_x > width
    
    #drop timestamp - outside coordinates
    df = df.loc[~(df['timestamp']>8000)]
    
    #drop all non fixations
    df = df.loc[~(df['sg_fix_all']==False)]
    
    #loop to draw event markers
    def xline(x, **kwargs):
        data = kwargs.pop("data")
        trialType = data['trialType'].values[0]
        TrialNum = data['TrialNum'].values[0]
        #check if first iaps or first pofa
        if TrialNum in first:
            #fixation onset
            time = events['fixation']
            plt.axvline(x=0, c='red', linewidth=2)
            ##annotation
            plt.text(s="Fixation Onset", x=(0 + 100), y=(xy[0] - 25), fontsize=12)
            #fixation offset
            time = events['fixation']
            plt.axvline(x=time, c='red', linewidth=2)
            ##annotation
            plt.text(s="Stimulus Offset", x=(time + 100), y=(xy[0] - 25), fontsize=12)
            #stimulus offset
            time = events['fixation'] + events['stimulus'][trialType]
            #line
            plt.axvline(x=time, c='red', linewidth=2)
            #annotation
            plt.text(s="Dotloc Offset", x=(time + 100), y=(xy[0] - 25), fontsize=12)
        
    #plot    
    pal = sns.color_palette("RdBu", n_colors=192)    
    sp = sns.FacetGrid(col="trialType", data=df, height=5, palette=pal, hue="TrialNum")
    sp.map(plt.scatter, "timestamp", "sg_x", alpha=0.3, s=10)
    sp.map_dataframe(xline, x="timestamp")
            
    #post
    ax = sp.axes.flatten()
    ax[0].set_ylabel("Gaze Coordinates (x)", fontdict={'fontsize':12})
    ax[0].set_title("trialType = IAPS", fontdict={'fontsize':12})
    ax[0].tick_params(labelsize=12)
    ax[1].set_title("trialType = POFA", fontdict={'fontsize':12})
    ax[1].tick_params(labelsize=12)
    
    #size
    sp.fig.set_size_inches(20,10)
    
    #save
    sp.savefig(path, dpi=300)
    
    #clear figure from memory
    plt.close()
    
    #--------finished
    #timestamp
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plt 

def boxplot(config,df,path=None,x=None,y=None,title=None,plots=None,cat='analysis'):
    """
    Creates boxplot using seaborn and pandas.
    
    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    path : :obj:`str`
        Path to save data.
    drift : :obj:`str`
        X-axis.
    drift : :obj:`str`
        Y-axis.
    title : :obj:`str`
        Plot title.
    plots : :obj:`dict`
        Dictionary of plots metadata.
    cat : :obj:`str`
        Type of plot.
    """ 
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('running boxplot(%s)'%(y), 'blue')

	# start
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    importlib.reload(plt); importlib.reload(sns)
    sns.set(style=config['style']['seaborn'], font_scale=1.5, font="Helvetica")
    sns.despine(offset=10, trim=True)
    short_ = config['metadata']['short']
    #if looking at dwell time
    if cat=='bias':
        #plot
        fig, (ax1, ax2) = plt.subplots(figsize=(10,10),ncols=2,sharey=True)
        
        #design
        ##for each unique value in x-axis variable, assign color
        cat_x = df[x].unique().tolist()
        color=['b','r','g','y','o']
        pal = dict(zip(cat_x, color))
        
        #create plot
        for ax, trialType in zip([ax1,ax2],['iaps','pofa']):
            df_ = df.loc[df['trialType'].isin([trialType])]
            sns.boxplot(x=x, y='.fitted', data=df_, ax=ax, palette=pal)
            sns.swarmplot(x=x, y='.fitted', data=df_, color=".25", ax=ax)
            ax.set_title(trialType)
            ax.set_xlabel(short_[x])
            if trialType == 'pofa':
                ax.set_ylabel('')
    
    #if looking at dwell time
    if cat=='dwell':
        #plot
        fig, (ax1, ax2) = plt.subplots(figsize=(10,10),ncols=2,sharey=True)
        
        #design
        ##for each unique value in x-axis variable, assign color
        cat_x = df[x].unique().tolist()
        color=['b','r','g','y','o']
        pal = dict(zip(cat_x, color))
        
        #create plot
        for ax, trialType in zip([ax1,ax2],['iaps','pofa']):
            df_ = df.loc[df['trialType'].isin([trialType])]
            sns.boxplot(x=x, y='.fitted', data=df_, ax=ax, palette=pal)
            sns.swarmplot(x=x, y='.fitted', data=df_, color=".25", ax=ax)
            ax.set_title(trialType)
            ax.set_xlabel(short_['aoi'])
            if trialType == 'pofa':
                ax.set_ylabel('')
    
    #if looking at variables like gaze_bias, dp_bias
    elif cat=='analysis':
        #plot
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(figsize=(19.2, 10.8),ncols=4,sharey=False)
        
        #design
        ##for each unique value in x-axis variable, assign color
        cat_x = df[x].unique().tolist()
        color=['b','r','g','y','o']
        pal = dict(zip(cat_x, color))
        
        #create plot
        for ax, itm in zip([ax1,ax2,ax3,ax4],y):
            #dp_bias
            sns.boxplot(x=x, y=itm, data=df, ax=ax, palette=pal)
            sns.swarmplot(x=x, y=itm, data=df, color=".25", ax=ax)
            ax.set(xlabel=x)
            
    #if looking at variables like race, gender, os        
    elif cat=='demographics':
        #layout
        sns.set(style=config['style']['seaborn'], palette="deep", font_scale=1.1, font="Helvetica")
        sns.despine(offset=10, trim=True)
        fig, (ax1,ax2,ax3,ax4) = plt.subplots(figsize=(20, 10),ncols=4,sharey=True)
        palette = sns.color_palette(["#e74c3c", "#3f51b5", "#2ecc71","#9b59b6","#ffb400"])
        #counter
        counter = 0
        for ax, x_, in zip([ax1,ax2,ax3,ax4],x):
            ##exclude smaller race categories for display purposees
            if x_=='gender':
                df_ = df.loc[~df['gender'].isin(["other"])].drop_duplicates(subset="participant",keep="first")
            ##exclude smaller race categories for display purposes
            elif x_=='race':
                df_ = df.loc[~df['race'].isin(["American Indian or Alaska Native","Two or more races",
                "Black or African American","None of the above"])].drop_duplicates(subset="participant",keep="first")
            else:
                df_ = df.drop_duplicates(subset="participant",keep="first")
            #plot
            sns.boxplot(x=x_, y=y, palette=palette, ax=ax, data = df_)
            #remove label
            if counter!=0:
                ax.set_ylabel('')
            #counter
            counter = counter + 1
    
        #layer
        sns.despine(offset=10, trim=True)
    
    #tight layout
    fig.tight_layout()
    
    ##save
    fig.savefig(path, dpi=300)
    
    #clear from memory
    plt.close(fig) 
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')

def cooks_plot(config, y, model, path, df):
    """Create cooks distance plot plot using seaborn, pandas, and statsmodel.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    path : :obj:`str`
        The directory path save the seaborn plot.
    y : :obj:`str`
        The predictor variable.
    model : :obj:`dict`
        statsmodel model.

    Returns
    -------
    res
        seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    console('running cooks_plot(%s)'%(y), 'blue')
    #-------plot
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)

    #plot design
    sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
    sns.despine(offset=10, trim=True)

    #plot
    fig, ax = plt.subplots(figsize=(10,10))
    
    # influence
    influence = model.get_influence()
    
    # cook's distance, from statsmodels internals
    ## c is the distance and p is p-value
    (c, p) = influence.cooks_distance
    
    #plot
    plt.stem(np.arange(len(c)), c, markerfmt=",")
    plt.xlabel('Observation Number')
    plt.ylabel('Cooks Distance')
    
    # annotations
    particpants = list(df['participant'])
    leverage_lines = np.flip(np.argsort(c), 0)[:3]
    for i in leverage_lines:
        plt.annotate(particpants[i], xy=(i, c[i]))

    #tight layout
    fig.tight_layout()
    
    #save
    fig.savefig(path, dpi=300)
    
    #clear figure from memory
    plt.close(fig)
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plt
   
def residual_plot(config, y, residuals, path):
    """Create probability plot using seaborn, pandas, and rpy2.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    y : :obj:`str`
        Predictor variable.
    residuals : :class:`pandas.DataFrame`
        Pandas dataframe of residuals vs fitted, qq data, and raw data.
    path : :obj:`str`
        The directory path save the seaborn plot.

    Returns
    -------
    lmp
        seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    
    console('running residual_plot(%s)'%(y), 'blue')
    #-------plot
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)
    
    #-------plot design
    sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
    sns.despine(offset=10, trim=True)
    #-------plot
    fig, ax = plt.subplots(figsize=(10,10))
    #residual vs fitted plot
    cmap=sns.color_palette(["#bd282f","#226db2"])
    sns.scatterplot(x=".fitted", y=".resid", data=residuals, ax=ax, hue="trialType", palette=cmap, legend=None, s=60, edgecolor=None)
    #loess smoothing curve
    sns.regplot(x=".fitted", y=".resid", data=residuals, ax=ax, lowess=True, scatter=False, color="blue")
    #horizontal line
    plt.axhline(0, zorder=1, color='black', linewidth=2.25)
    #annotations
    ##max
    max_ = residuals[residuals['.resid']==residuals['.resid'].max()]
    max_x = max_['.fitted'].iloc[0]
    max_y = max_['.resid'].iloc[0]
    max_subject = max_['participant'].iloc[0]
    plt.text(max_x, max_y, max_subject, color='black', weight='semibold')
    ##min
    min_ = residuals[residuals['.resid']==residuals['.resid'].min()]
    min_x = min_['.fitted'].iloc[0]
    min_y = min_['.resid'].iloc[0]
    min_subject = min_['participant'].iloc[0]
    plt.text(min_x, min_y, min_subject, color='black', weight='semibold')

    #-------plot design
    #label
    plt.xlabel('.fitted')
    plt.ylabel('.resid')
    #tight layout
    fig.tight_layout()
    
    #-------save
    fig.savefig(path, dpi=300)
    
    #-------clear figure from memory
    plt.close(fig)
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plt

def qq_plot(config, y, residuals, path):
    """Create probability plot using seaborn, pandas, and rpy2.

    Parameters
    ----------
    config : :obj:`dict`
        Configuration data.
    y : :obj:`str`
        Predictor variable.
    residuals : :class:`pandas.DataFrame`
        Pandas dataframe of residuals vs fitted, qq data, and raw data.
    path : :obj:`str`
        The directory path save the seaborn plot.

    Returns
    -------
    lmp
        seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('running qq_plot(%s)'%(y), 'blue')
    #-------plot
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)

    #plot design
    sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
    sns.despine(offset=10, trim=True)
    #plot
    fig, ax = plt.subplots(figsize=(10,10))
    #qq residuals
    cmap=sns.color_palette(["#bd282f","#226db2"])
    sns.scatterplot(x="theoretical", y=".resid", data=residuals, ax=ax, hue="trialType", palette=cmap, legend=None, s=60, edgecolor=None)
    sns.regplot(x="theoretical", y=".resid", data=residuals, ax=ax, scatter=None, color="black", ci=None)
    #label
    plt.xlabel('Theoretical Quantiles')
    plt.ylabel('Sample Quantiles')
    #annotations
    ##max
    max_ = residuals[residuals['.resid']==residuals['.resid'].max()]
    max_x = max_['theoretical'].iloc[0]
    max_y = max_['.resid'].iloc[0]
    max_subject = max_['participant'].iloc[0]
    plt.text(max_x, max_y, max_subject, color='black', weight='semibold')
    ##min
    min_ = residuals[residuals['.resid']==residuals['.resid'].min()]
    min_x = min_['theoretical'].iloc[0]
    min_y = min_['.resid'].iloc[0]
    min_subject = min_['participant'].iloc[0]
    plt.text(min_x, min_y, min_subject, color='black', weight='semibold')
    
    #tight layout
    fig.tight_layout()
    
    #save
    fig.savefig(path, dpi=300)
    
    #clear figure from memory
    plt.close(fig)
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return plt

def logit_plot(config,df,path,param):
    """Create logistic regression plot using seaborn and pandas

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    path : :obj:`str`
        The directory path save the seaborn plot.
    param : :obj:`dict`
        x, y, groupby parameters.

    Returns
    -------
    lmp
        seaborn plot.
    """
    #----initiate fonts
    __font__()
    
    #----for timestamp
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('processing.logit_plot()', 'blue')
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    importlib.reload(plt); importlib.reload(sns)
    
    #plot design
    p=param
    sns.set(style=config['style']['seaborn'], font_scale=1.1, font="Helvetica")
    sns.despine(offset=10, trim=True)
    console('running logit_plot(), (y=%s, x=%s)'%(p['y'],p['x']), 'blue')

    #plot
    fig, (ax1,ax2) = plt.subplots(figsize=(19.2, 10.8), ncols=2, sharey=True)
    #iaps
    type_ = 'iaps'
    df_ = df.loc[(df['trialType_']==type_)]
    lmp = sns.regplot(x=p['x'], y=p['y'], data=df_, logistic=True, y_jitter=.02, ax=ax1, color="#5874a6")
    ax1.set(xlabel='trial type', title='trialType=%s'%(type_))
    #pofa
    type_ = 'pofa'
    df_ = df.loc[(df['trialType_']==type_)]
    lmp = sns.regplot(x=p['x'], y=p['y'], data=df_, logistic=True, y_jitter=.02, ax=ax2, color="#5d9f6c")
    ax2.set(xlabel='trial type', title='trialType=%s'%(type_))

    #tight layout
    fig.tight_layout()
    
    #save
    fig.savefig(path, dpi=300)
    
    #clear figure from memory
    plt.close(fig)
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return lmp

def html(config, df=None, raw_data=None, name=None, path=None, plots=None, source=None, title=None, intro=None, footnote=None, script="", **kwargs):
    """
    Create HTML output.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        Pandas dataframe of analysis results data.
    raw_data : :class:`pandas.DataFrame`
        Pandas dataframe of raw data.
    name : :obj:`str`
        (py::`if source is logit`) The name of csv file created.
    path : :obj:`str`
        The directory path of the html file.
    plots : :obj:`dict`
        If generating seaborn images, the list of plots used.
    source : :obj:`str`
        The type of data being recieved.
    trial : :obj:`str`
        (If Bokeh) Trial Number.
    session : :obj:`str`
        (If Bokeh) Session Number.
    bokeh_type : :obj:`str`
        (If Bokeh) Control directory location. If trial, create trial plots.
    title : :obj:`str`
        The title of the table or figure.
    intro : :obj:`str`
        The introduction of the group of figures or tables.
    footnote : :obj:`str`
        The footnote of the table or figure.
    metadata : :obj:`dict`
        Additional data to be included.
    **kwargs : :obj:`str`, :obj:`int`, or :obj:`None`, optional
        Additional properties, relevent for specific content types. Here's a list of available properties:
            
        .. list-table::
           :class: kwargs
           :widths: 25 50
           :header-rows: 1
           
           * - Property
             - Description
           * - **short**, **long** : :obj:`str`
             - Short (aoi) and long form (Area of Interest) label of html page. This is primarily used for constructing metadata tags in html.
           * - **display** : :obj:`str`
             - (For bokeh) The type of calibration/validation display.
           * - **trial** : :obj:`str`
             - (For bokeh) The trial number for the eyetracking task.
           * - **session** : :obj:`int`
             - (For bokeh) The session number for the eyetracking task.
           * - **day** : :obj:`str`
             - (For bokeh) The day the eyetracking task was run.

    Returns
    -------
    html : :obj:`str`
        String of html code.
    """
    #----initiate fonts
    __font__()
    
    #timestamp
    date = datetime.datetime.now().replace(microsecond=0).strftime('%Y-%m-%d %H:%M:%S')
    _t0 = datetime.datetime.now()
    _f = debug(message='t', source="timestamp")
    console('running plot.html()', 'blue')
    
    #---copy csv, imhr, js files to each model
    if ((source == "logit") or (source == "onset") or (source == "anova")):
        #copy files
        for folder in ['css','imhr','js']:
            source_ = config['path']['output'] + "/analysis/html/" + folder + "/"
            dest_ = os.path.dirname(path) + "/" + folder + "/"
            dir_util.copy_tree(src=source_, dst=dest_)
    
    #----if creating html file from bokeh
    if ((source=='bokeh') or (source=='methods')):
        #get short and long names
        short = kwargs['short'] if 'short' in kwargs else source
        long = kwargs['long'] if 'long' in kwargs else source
        
        #if trial plot
        if (source=='bokeh'):
            display =  kwargs['display'] if 'display' in kwargs else None
            trial =  kwargs['trial'] if 'trial' in kwargs else None
            session =  kwargs['session'] if 'session' in kwargs else None
            day =  kwargs['day'] if 'day' in kwargs else None
            if ((display=='trial') or (display=='calibration') or (display=='validation')):
                parent='../'
            else:
                parent=''
        else:
            parent=''
            
        #build header
        link = []; css = []; js = []
        link.append('<title>IMHR [webgazer] %s</title>'%(long))
        link.append('<meta name="title" content="mdl[webgazer]: %s">'%(long))
        link.append('<meta name="description" content="Results from our feasibility study for in-browser eyetracking.">')
        link.append('<meta name="author" content="Semeon Risom">')
        link.append('<meta name="email" content="semeon.risom@gmail.com">')
        link.append('<meta name="robots" content="index,follow">')
        link.append('<meta name="AdsBot-Google" content="noindex">')
        link.append('<!--size-->')
        link.append('<meta name="viewport" content="width=device-width, initial-scale=1.0" charset="utf-8"/>')
        link.append('<!--no cache-->')
        link.append('<meta http-equiv="Cache-Control" content="no-cache, no-store, must-revalidate"/>')
        link.append('<meta http-equiv="Pragma" content="no-cache"/>')
        link.append('<meta http-equiv="Expires" content="0"/>')
        link.append('<!--icons-->')
        link.append('<link rel="shortcut icon" href="'+parent+'imhr/favicon/favicon.ico">')
        link.append('<link rel="icon" sizes="16x16 32x32 64x64" href="'+parent+'imhr/favicon/favicon.ico">')
        link.append('<link rel="icon" type="image/png" sizes="196x196" href="'+parent+'imhr/favicon/favicon-192.png">')
        link.append('<link rel="icon" type="image/png" sizes="160x160" href="'+parent+'imhr/favicon/favicon-160.png">')
        link.append('<link rel="icon" type="image/png" sizes="96x96" href="'+parent+'imhr/favicon/favicon-96.png">')
        link.append('<link rel="icon" type="image/png" sizes="64x64" href="'+parent+'imhr/favicon/favicon-64.png">')
        link.append('<link rel="icon" type="image/png" sizes="32x32" href="'+parent+'imhr/favicon/favicon-32.png">')
        link.append('<link rel="icon" type="image/png" sizes="16x16" href="'+parent+'imhr/favicon/favicon-16.png">')
        link.append('<link rel="apple-touch-icon" href="'+parent+'imhr/favicon/favicon-57.png">')
        link.append('<link rel="apple-touch-icon" sizes="114x114" href="'+parent+'imhr/favicon/favicon-114.png">')
        link.append('<link rel="apple-touch-icon" sizes="72x72" href="'+parent+'imhr/favicon/favicon-72.png">')
        link.append('<link rel="apple-touch-icon" sizes="144x144" href="'+parent+'imhr/favicon/favicon-144.png">')
        link.append('<link rel="apple-touch-icon" sizes="60x60" href="'+parent+'imhr/favicon/favicon-60.png">')
        link.append('<link rel="apple-touch-icon" sizes="120x120" href="'+parent+'imhr/favicon/favicon-120.png">')
        link.append('<link rel="apple-touch-icon" sizes="76x76" href="'+parent+'imhr/favicon/favicon-76.png">')
        link.append('<link rel="apple-touch-icon" sizes="152x152" href="'+parent+'imhr/favicon/favicon-152.png">')
        link.append('<link rel="apple-touch-icon" sizes="180x180" href="'+parent+'imhr/favicon/favicon-180.png">')
        link.append('<meta name="msapplication-TileColor" content="#FFFFFF">')
        link.append('<meta name="msapplication-TileImage" content="'+parent+'imhr/favicon/favicon-144.png">')
        link.append('<meta name="msapplication-config" content="'+parent+'docs/favicon/browserconfig.xml">')
        css.append('<link rel="stylesheet" type="text/css" href="'+parent+'css/user.css">')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/jquery/jquery-3.2.1.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/user.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/lodash.min.js"></script>')
        css.append('<!--bootstrap-->')
        css.append('<link rel="stylesheet" type="text/css" href="'+parent+'css/bootstrap/bootstrap.css">')
        css.append('<!--bokeh-->')
        css.append('<link rel="stylesheet" type="text/css" href="'+parent+'css/bokeh/bokeh-1.0.4.min.css">')
        css.append('<link rel="stylesheet" type="text/css" href="'+parent+'css/bokeh/bokeh-widgets-1.0.4.min.css">')
        css.append('<link rel="stylesheet" type="text/css" href="'+parent+'css/bokeh/bokeh-tables-1.0.4.min.css">')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bokeh/bokeh-1.0.4.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bokeh/bokeh-widgets-1.0.4.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bokeh/bokeh-tables-1.0.4.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bokeh/bokeh-gl-1.0.4.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bokeh/bokeh-api-1.0.4.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/popper.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="'+parent+'js/bootstrap/bootstrap.js"></script>')
        
        # finish header
        head = ('\n\t'.join(map(str, link))) + ('\n\t'.join(map(str, css))) + ('\n\t'.join(map(str, js)))
        
        #body, footer
        if source=='bokeh':
            #build body
            body = ['<div class="container" style="">',
                        '<div class="container bokeh" style="">',
                            '<div class="figure %s" trial-id="%s" session-id="%s" day-id="%s" type-id="%s">\
                            '%(display, trial, session, day, display),
                                '<div class="title"><b>Figure 1.</b> <a>%s</a></div>'%(title),
                                '<div class="date"> Last Updated: ' + '%s'%(date) + '</div>',
                                plots,
                                '<div class="footnote">%s</div>'%(footnote),
                            '</div>',
                        '</div>',
                    '</div>']
            # finish body
            body = ('\n\t'.join(map(str, body)))
            
            #build footer
            if ((display=='trial') or (display=='calibration') or (display=='validation')):
                foot = ['<script>',
                            '$(document).ready(function() {getinput();});',
                        '</script>']
                # finish footer
                foot = ('\n\t'.join(map(str, foot)))
            else:
                foot = ''
                
        elif source=='methods':
            #build body
            body = ['<div class="container" style="">',
                        '<div class="dashboard-main">',
                            '<div class="table-container">',
                                '<div class="title"><a><b>%s</b></a></div>'%(title) ,
                                plots,
                            '</div>',
                        '</div>',
                    '</div>']
            # finish body
            body = ('\n\t'.join(map(str, body)))
            #build footer
            foot = ''
        
        #build html
        html = ['<html short="%s" long="%s">'%(short, long),
                    '<head>',
                        head,
                    '</head>',
                    '<body id="body" class="%s" style="margin: 0px; background-color: #434e54 !important;">'%(display),
                        body,
                    '</body>',
                    '<footer>',
                        foot,
                    '</footer>',
                '</html>']
        # finish html
        html = ('\n'.join(map(str, html)))
        
    #all other html
    else:
        #get short and long names
        short = kwargs['short'] if 'short' in kwargs else source
        long = kwargs['long'] if 'long' in kwargs else source
        var = kwargs['var'] if 'var' in kwargs else source
        
        #build header
        link = []; css = []; js = []
        link.append('<title>IMHR [webgazer] %s</title>'%(long))
        link.append('<meta name="title" content="mdl[webgazer]: %s">'%(long))
        link.append('<meta name="description" content="Results from our feasibility study for in-browser eyetracking.">')
        link.append('<meta name="author" content="Semeon Risom">')
        link.append('<meta name="email" content="semeon.risom@gmail.com">')
        link.append('<meta name="robots" content="index,follow">')
        link.append('<meta name="AdsBot-Google" content="noindex">')
        link.append('<!--size-->')
        link.append('<meta name="viewport" content="width=device-width, initial-scale=1.0" charset="utf-8"/>')
        link.append('<!--no cache-->')
        link.append('<meta http-equiv="Cache-Control" content="no-cache, no-store, must-revalidate"/>')
        link.append('<meta http-equiv="Pragma" content="no-cache"/>')
        link.append('<meta http-equiv="Expires" content="0"/>')
        link.append('<meta name="msapplication-TileColor" content="#FFFFFF">')
        link.append('<meta name="msapplication-TileImage" content="imhr/favicon/favicon-144.png">')
        link.append('<meta name="msapplication-config" content="docs/favicon/browserconfig.xml">')
        link.append('<!--icons-->')
        link.append('<link rel="shortcut icon" href="imhr/favicon/favicon.ico">')
        link.append('<link rel="icon" sizes="16x16 32x32 64x64" href="imhr/favicon/favicon.ico">')
        link.append('<link rel="icon" type="image/png" sizes="196x196" href="imhr/favicon/favicon-192.png">')
        link.append('<link rel="icon" type="image/png" sizes="160x160" href="imhr/favicon/favicon-160.png">')
        link.append('<link rel="icon" type="image/png" sizes="96x96" href="imhr/favicon/favicon-96.png">')
        link.append('<link rel="icon" type="image/png" sizes="64x64" href="imhr/favicon/favicon-64.png">')
        link.append('<link rel="icon" type="image/png" sizes="32x32" href="imhr/favicon/favicon-32.png">')
        link.append('<link rel="icon" type="image/png" sizes="16x16" href="imhr/favicon/favicon-16.png">')
        link.append('<link rel="apple-touch-icon" href="imhr/favicon/favicon-57.png">')
        link.append('<link rel="apple-touch-icon" sizes="114x114" href="imhr/favicon/favicon-114.png">')
        link.append('<link rel="apple-touch-icon" sizes="72x72" href="imhr/favicon/favicon-72.png">')
        link.append('<link rel="apple-touch-icon" sizes="144x144" href="imhr/favicon/favicon-144.png">')
        link.append('<link rel="apple-touch-icon" sizes="60x60" href="imhr/favicon/favicon-60.png">')
        link.append('<link rel="apple-touch-icon" sizes="120x120" href="imhr/favicon/favicon-120.png">')
        link.append('<link rel="apple-touch-icon" sizes="76x76" href="imhr/favicon/favicon-76.png">')
        link.append('<link rel="apple-touch-icon" sizes="152x152" href="imhr/favicon/favicon-152.png">')
        link.append('<link rel="apple-touch-icon" sizes="180x180" href="imhr/favicon/favicon-180.png">')
        js.append('<!--all-->')
        js.append('<script type="text/javascript" language="javascript" src="js/jquery/jquery-3.2.1.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/lodash.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/bootstrap/bootstrap.bundle.js"></script>')
        link.append('<!--dataTables-->')
        css.append('<link rel="stylesheet" type="text/css" href="css/bootstrap/bootstrap.min.css">')
        css.append('<link rel="stylesheet" type="text/css" href="css/datatables/dataTables.bootstrap.min.css">')
        css.append('<link rel="stylesheet" type="text/css" href="css/datatables/responsive.bootstrap.min.css">')
        css.append('<link rel="stylesheet" type="text/css" href="css/datatables/buttons.bootstrap.min.css">')
        js.append('<!--user-->')
        link.append('<link rel="stylesheet" type="text/css" href="css/user.css">')
        js.append('<!--dataTables-->')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/jquery.dataTables.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/dataTables.rowsGroup.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/dataTables.rowGroup.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/dataTables.bootstrap.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/dataTables.buttons.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/buttons.bootstrap.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/dataTables.responsive.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/responsive.bootstrap.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/buttons.html5.min.js"></script>')
        js.append('<script type="text/javascript" language="javascript" src="js/datatables/jszip.min.js"></script>')
        js.append('<!--user-->')
        js.append('<script type="text/javascript" language="javascript" src="js/user.js"></script>')
        js.append('<script>source="%s"</script>'%(source))
        # additional libraries
        if ((source == "logit") or (source == "onset") or (source == "anova")):
            link.append('<!--prism-->')
            css.append('<link rel="stylesheet" type="text/css" href="css/prism/prism.css">')
            css.append('<link rel="stylesheet" type="text/css" href="css/prism/prism-line-numbers.css">')
            css.append('<script type="text/javascript" language="javascript" src="js/prism/prism.js"></script>')
            css.append('<script type="text/javascript" language="javascript" src="js/prism/prism-r.js"></script>')
            css.append('<script type="text/javascript" language="javascript" src="js/prism/prism-python.js"></script>')
            css.append('<script type="text/javascript" language="javascript" src="js/prism/prism-line-numbers.js"></script>')
            css.append('<script type="text/javascript" language="javascript" src="js/prism/prism-normalize-whitespace.js"></script>')
        
        # finish header
        head = ('\n\t'.join(map(str, link))) + ('\n\t'.join(map(str, css))) + ('\n\t'.join(map(str, js)))
        
        #if displaying seaborn plots
        if source == "plots":
            html_plots = []
            #for each plot
            for idx, p in enumerate(plots):
                html_p = ['<div class="figure" style="max-width:800px;">',
                            #title
                            '<div class="title"><b>Figure %s.</b> %s</div>'%((idx+1), p['title']),
                            #date
                            '<div class="date"> Last Updated: ' + '%s'%(date) + '</div>',
                            #figure
                            '<a href="img\\' + p['file'] + '">\n<img class="figure" src="img\\' + p['file'] + '">\n</a>',
                            #footnote
                            '<div class="footnote">%s</div>'%(p['footnote']),
                        '</div>']
                html_plots.append(('\n\t'.join(map(str, html_p))))
                
            #build body
            ##add none as class only if there is no intro
            if intro!=None:
                intro_class = ''
            else:
                intro_class = None
                
            body = ['<div class="container" style="">',
                        '<div class="container-figure %s" style="">'%(display),
                            '<div class="intro %s">%s</div>'%(intro_class, intro),
                                (''.join(map(str, html_plots))),
                            '</div>',
                        '</div>',
                    '</div>']   
            body = ('\n\t'.join(map(str, body)))
            
            #build html
            html = ['<html>',
                    '<head>',
                        head,
                    '</head>',
                    '<body id="body">',
                        body,
                    '</body>',
                    '</html>']
            html = ('\n' .join(map(str, html)))
            
            #save
            with open(path,'w') as html_:
                html_.write(html)
                
            #--------finished
            console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
            return html
        
        #if displaying dataframe tables
        else:         
            #build body
            ##save table to html for datatable
            if ((source == "demographic") or (source == "device") or (source == "task") or (source == "summary")):
                #allow word wrapping
                if(source == "demographic") or (source == "task"):
                    wrap = "wrap"
                else:
                    wrap = "nowrap"
                # save table to csv for downloading
                ## check if paths exist
                p_ = config['path']['output'] + '/analysis/html/csv/'
                if not os.path.exists(p_):
                    os.makedirs(p_)
                df.to_csv(p_ + '%s.csv'%(name), index=False)
                #get table
                table = df.to_html(index=False, index_names=True).replace('<table border="1" class="dataframe">',
                '<table id="table" class="table '+source+' table-striped table-bordered \
                hover dt-responsive '+wrap+'" cellspacing="0" width="100%">').replace('&gt;','>').replace('&lt;','<')
            elif (source == "definitions"):
                #wordwrap
                wrap = "nowrap"
                #prevent trucating strings
                width = pd.get_option('display.max_colwidth')
                pd.set_option('display.max_colwidth', -1)
                ##save table to csv for downloading
                df.to_csv(config['path']['output'] + '/analysis/html/csv/%s.csv'%(name), index=False)
                #get table
                table = df.to_html(index=False, index_names=True).replace('<table border="1" class="dataframe">',
                '<table id="table" class="table '+source+' table-striped table-bordered \
                hover dt-responsive '+wrap+'" cellspacing="0" width="100%">')   
                #reset
                pd.set_option('display.max_colwidth', width)
            elif ((source == "logit") or (source == "onset") or (source == "anova")):
                #prevent trucating strings
                width = pd.get_option('display.max_colwidth')
                pd.set_option('display.max_colwidth', -1)
                ##save table to csv for downloading #here saving raw data for future analysis
                raw_data.to_csv(config['path']['output'] + '/analysis/html/csv/%s.csv'%(name), index=False)
                #get table
                table = df.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
                '<table id="table" data-file="'+ name +'" class="table '+source+' table-striped table-bordered \
                hover dt-responsive nowrap" cellspacing="0" width="100%">')
                #reset
                pd.set_option('display.max_colwidth', width)
            else:
                ##check if paths exist
                p_ = config['path']['output'] + '/analysis/html/csv/'
                if not os.path.exists(p_):
                    os.makedirs(p_)
                # save table to csv for downloading
                df.to_csv(p_ + '%s.csv'%(name), index=False)
                #get table
                table = df.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
                '<table id="table" data-file="'+ name +'" class="table '+source+' table-striped table-bordered \
                hover dt-responsive nowrap" cellspacing="0" width="100%">')
      
            #add images to bottom of model
            if ((source == "logit") or (source == "onset") or (source == "anova")):
                html_plots = []
                #for each plot
                if plots != None:
                    fig_num = 0
                    for idx, p in enumerate(plots):
                        if (source == "anova"):
                            #if supplimentary table
                            if (p['type'] == 'table'):
                                if 'anchor' in p:
                                    anchor = p['anchor']
                                _table = p['df']
                                #prevent trucating strings
                                width = pd.get_option('display.max_colwidth')
                                pd.set_option('display.max_colwidth', -1)
                                #convert df to html table
                                _table = _table.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
                                '<table id="table2" data-file="'+ name +'" class="table '+source+' table-striped table-bordered \
                                hover dt-responsive nowrap"\ cellspacing="0" width="100%">')
                                #create table
                                html_p = [
                                    '<div class="figure %s" id="%s">'%(source, anchor),
                                        '<div class="title">%s</div>'%(p['title']),
                                        '<div class="date"> Last Updated: ' + '%s'%(date) + '</div>',
                                        '<div class="dataTables_wrapper form-inline dt-bootstrap">',
                                            '<!--script-->' + '\n',
                                            _table,
                                        '</div>',
                                        '<div class="footnote">%s</div>'%(p['footnote']),
                                    '</div>']
                                html_p = ('\n\t'.join(map(str, html_p)))
                                #reset
                                pd.set_option('display.max_colwidth', width)
                            #else is plot
                            else:
                                fig_num = fig_num + 1
                                if 'anchor' in p:
                                    anchor = p['anchor']  
                                else: 
                                    anchor = ""
                                html_p = ['<div class="figure %s" id="%s">'%(source, anchor),
                                              #title
                                              '<div class="title"><b>Figure %d.</b> %s</div>'%(fig_num, p['title']),
                                              #date
                                              '<div class="date"> Last Updated: ' + '%s'%(date) + '</div>',
                                              #figure
                                              '<a class="img" href="img\\'+p['file']+'">\n<img class="img" src="img\\'+p['file']+'">\n</a>',
                                              #footnote
                                              '<div class="footnote">%s</div>'%(p['footnote']),
                                        '</div>']
                        #else is plot
                        else:
                            fig_num = fig_num + 1
                            if 'anchor' in p:
                                anchor = p['anchor']  
                            else: 
                                anchor = ""
                            html_p = ['<div class="figure %s" id="%s">'%(source, anchor),
                                        #title
                                        '<div class="title"><b>Figure %d.</b> %s</div>'%(fig_num, p['title']),
                                        #date
                                        '<div class="date"> Last Updated: ' + '%s'%(date) + '</div>',
                                        #figure
                                        '<a class="img" href="img\\'+p['file']+'">\n<img class="img" src="img\\'+p['file']+'">\n</a>',
                                        #footnote
                                        '<div class="footnote">%s</div>'%(p['footnote']),
                                    '</div>']
                        
                        #----append to html plots 
                        html_plots.append(('\n\t'.join(map(str, html_p))))
                    
                html_plots = '\n\t'.join(map(str, html_plots))
            else:
                html_plots = ""
                
            ##body
            body = ['<div class="container-large %s" style="">'%(source),
                        '<div class="container" style="">',
                            '<div class="dashboard-main">',
                                '<div class="table-container %s">'%(source),
                                    '<div class="title">'+'%s'%(title)+'</div>',
                                    '<div class="date"> Last Updated: '+'%s'%(date),
                                        '<span class="link" id="code">code</span>',
                                        '<span class="link" id="csv">csv</span>',
                                    '</div>',
                                '<div class="dataTables_wrapper form-inline dt-bootstrap">',
                                    '<!--script-->',
                                    table,
                                '</div>',
                                '<div class="footnote">' + '\n' + '%s'%(footnote) + '</div>',
                                    html_plots,
                                '</div>',
                            '</div>',
                        '</div>',
                    '</div>']
            #replace script within body, then join as string
            if ((source == "logit") or (source == "onset") or (source == "anova")):
                body = [script if x==(                '<!--script-->') else x for x in body]
            # join
            body = ('\n\t'.join(map(str, body)))

            #build footer
            if ((source == "logit") or (source == "onset") or (source == "anova")):
                foot = ['<script>',
                            '$(document).ready(function(){',
                                'getLogitTable();',
                                'additionalTable();',
                                'popover();',
                                'codeButton();',
                            '})',
                        '</script>']
            elif (source == "task"):
                foot = ['<script>',
                            '$(document).ready(function(){',
                                'getLogitTable();',
                                'popover();',
                            '})',
                        '</script>']   
            else:
                foot = ['<script>',
                            '$(document).ready(function(){',
                                'getLogitTable();',
                            '})',
                        '</script>']
            foot = ('\n\t'.join(map(str, foot)))
            
            #build html
            html = ['<html short="%s" long="%s">'%(short, long),
                    '<head>',
                        head,
                    '</head>',
                    '<body id="body">',
                        body,
                    '</body>',
                    '<foot>',
                        foot,
                    '</foot>',
                    '</html>']
            html = ('\n'.join(map(str, html)))
                
    #----save
    # check if paths exist
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    # tidy html
    #html = re.sub("\s\s+", " ", html) #multiple spaces
    # save
    with open(path,'w') as html_:
        html_.write(html)
    console('saved html file at %s'%(path), 'blue')
    
    #--------finished
    console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
    return html



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
| @purpose: Classification of eyetracking data for mdl.r33.procesing.   
| @date: Created on Sat May 1 15:12:38 2019   
| @author: Semeon Risom   
| @email: semeon.risom@gmail.com   
| @url: https://semeon.io/d/R33-analysis  
"""

# available functions
__all__ = ['Classify']

# required external library
__required__ = ['numpy','pandas','nslr']

from pdb import set_trace as breakpoint
import numpy as np
import pandas as pd

# local libraries
if __name__ == "__main__":
	from .. import settings

class Classify():
    """Analysis methods for mdl.processing.preprocesing."""
    def __init__(self, isLibrary=False):
        """
		Initiate the mdl.r33.Classify module.

        Parameters
        ----------
        isLibrary : :obj:`bool`
            Check if required libraries are available. Default `False`.
		"""
        #check libraries
        if isLibrary:
            settings.library(__required__)

    def VisualAngle(self ,g_x, g_y, config):
        """
        Convert pixel eye-coordinates to visual angle.

        Parameters
        ----------
        g_x,g_y : :class:`numpy.ndarray`
            List of gaze coordinates.
        drift : :obj:`dict`
            Counter of drift correct runs.
		config : :class:`dict`
            Configuration data for data analysis. i.e. trial number, location.

        Notes
        -----
        Stimulus positions (g_x,g_y) are defined in x and y pixel units, 
        with the origin (0,0) being at the **center** of the display, as to match
        the PsychoPy pix unit coord type.  

        The pix2deg method is vectorized, meaning that is will perform the 
        pixel to angle calculations on all elements of the provided pixel
        position numpy arrays in one numpy call.  

        The convertion process can use either a fixed eye to calibration 
        plane distance, or a numpy array of eye distances passed as
        eye_distance_mm. In this case the eye distance array must be the same
        length as g_x, g_y arrays.  
        """

        d_x=config['monitor.cm'][0]
        d_y=config['monitor.cm'][1]
        r_x=config['resolution.px'][0]
        r_y=config['resolution.px'][1]
        dist=config['distance']
        mmpp_x = d_x/r_x
        mmpp_y = d_y/r_y

        x_mm = mmpp_x * g_x
        y_mm = mmpp_y * g_y

        breakpoint() #TODO!

        Ah = np.arctan2(x_mm, np.hypot(dist, y_mm))
        Av = np.arctan2(y_mm, np.hypot(dist, x_mm))

        return np.rad2deg(Ah), np.rad2deg(Av)   

    def Velocity(self, time, config, d_x, d_y=None):
        """
        Calculate the instantaneous velocity (degrees / second) for data points in 
        d_x and (optionally) d_y, using the time numpy array for time delta information.

        Parameters
        ----------
        time : :class:`numpy.ndarray`
            Timestamp of each coordinate.
        d_x,d_y : :class:`numpy.ndarray`
            List of gaze coordinates.
		config : :class:`dict`
            Configuration data for data analysis. i.e. trial number, location.

        Notes
        -----
        Numpy arrays time, d_x, and d_y must all be 1D arrays of the same
        length.

        If both d_x and d_y are provided, then the euclidian distance
        between each set of points is calculated and used in the velocity calculation.

        Time must be in seconds.msec units, while d_x and d_y are expected
        to be in visual degrees. If the position traces are in pixel coordinate space,
        use the VisualAngleCalc class to convert the data into degrees.
        """
        if d_y is None:
            data=d_x
        else:
            data=np.sqrt(d_x*d_x+d_y*d_y)

        velocity_between = (data[1:]-data[:-1])/(time[1:]-time[:-1])
        velocity = (velocity_between[1:]+velocity_between[:-1])/2.0
        return velocity

    def Acceleration(self, time, data_x, data_y=None):
        """
        Calculate the acceleration (deg/sec/sec) for data points in 
        d_x and (optionally) d_y, using the time numpy array for 
        time delta information.

        Parameters
        ----------
        time : :class:`numpy.ndarray`
            Timestamp of each coordinate.
        d_x,d_y : :class:`numpy.ndarray`
            List of gaze coordinates.
		config : :class:`dict`
            Configuration data for data analysis. i.e. trial number, location.
        
		"""
        velocity=Velocity(time,data_x,data_y)
        accel = Velocity(time[1:-1],velocity)
        return accel

    def savitzky_golay(self, y, window_size, order, deriv=0, rate=1):
        """
        Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
        The Savitzky-Golay filter removes high frequency noise from data.
        It has the advantage of preserving the original shape and
        features of the signal better than other types of filtering
        approaches, such as moving averages techniques.

        Parameters
        ----------
        y : :class:`numpy.ndarray`, shape (N,)
            the values of the time history of the signal.
        window_size : :obj:`int`
            the length of the window. Must be an odd integer number.
        order : :obj:`int`
            the order of the polynomial used in the filtering.
            Must be less then `window_size` - 1.
        deriv : :obj:`int`
            the order of the derivative to compute (default = 0 means only smoothing)

        Returns
        -------
        ys : :class:`numpy.ndarray`, shape (N)
            the smoothed signal (or it's n-th derivative).

        Notes
        -----
        The Savitzky-Golay is a type of low-pass filter, particularly
        suited for smoothing noisy data. The main idea behind this
        approach is to make for each point a least-square fit with a
        polynomial of high order over a odd-sized window centered at
        the point.
		For more information, see: http://wiki.scipy.org/Cookbook/SavitzkyGolay.

        Examples
        --------
        >>> t = np.linspace(-4, 4, 500)
        >>> y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
        >>> ysg = savitzky_golay(y, window_size=31, order=4)
        >>> import matplotlib.pyplot as plt
        >>> plt.plot(t, y, label='Noisy signal')
        >>> plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
        >>> plt.plot(t, ysg, 'r', label='Filtered signal')
        >>> plt.legend()
        >>> plt.show()

        References
        ----------
        .. [1] A. Savitzky, Golay, M. (1964). Smoothing and Differentiation of Data by Simplified Least Squares Procedures. 
			Analytical Chemistry. 36(8), pp 1627-1639.

        .. [2] S.A. Teukolsky, W.T. Vetterling, B.P. Flannery Numerical Recipes 3rd Edition: The Art of Scientific 
			Computing. W.H. Press,Cambridge University Press ISBN-13: 9780521880688.
        """
        import numpy as np
        from math import factorial

        try:
            window_size = np.abs(np.int(window_size))
            order = np.abs(np.int(order))
        except ValueError:
            raise ValueError("window_size and order have to be of type int")
        if window_size % 2 != 1 or window_size < 1:
            raise TypeError("window_size size must be a positive odd number")
        if window_size < order + 2:
            raise TypeError("window_size is too small for the polynomials order")
        order_range = range(order+1)
        half_window = (window_size -1) // 2
        # precompute coefficients
        b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
        m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
        # pad the signal at the extremes with
        # values taken from the signal itself
        firstvals = y[0] - np.abs(y[1:half_window+1][::-1] - y[0])
        #breakpoint() #TODO!
        if isinstance(y, pd.core.series.Series):
            y_m1 = y.iloc[-1]
        else:
            y_m1 = y[-1]
        lastvals = y_m1 + np.abs(y[-half_window-1:-1][::-1] - y_m1)
        y = np.concatenate((firstvals, y, lastvals))
        return np.convolve( m[::-1], y, mode='valid')

    def ivt(self,data, v_threshold, config):
        """
		Identification with Velocity Threshold.
        
		
		In the I-VT model, the velocity value is computed for every eye position sample. 
        The velocity value is then compared to the threshold. If the sampled velocity is less 
        than the threshold, the corresponding eye-position sample is marked as part of a fixation, 
        otherwise it is marked as a part of a saccade.

        Parameters
        ----------
        data : :class:`numpy.ndarray`, shape (N,)
            the smoothed signal (or it's n-th derivative).
        v_threshold : :obj:`str`
            Velocity threshold in pix/sec.
		config : :class:`dict`
            Configuration data for data analysis. i.e. trial number, location.

        Returns
        -------
        ys : :class:`numpy.ndarray`, shape (N,)
            the smoothed signal (or it's n-th derivative).

        Notes
        -----
        From https://github.com/ecekt/eyegaze.
        Formula from: https://dl.acm.org/citation.cfm?id=355028

        """
        l_time = data['timestamp']
   
        ts = []

        for t in l_time:
            ts.append(float(t)/1000.0)

        Xs = data['x']
        Ys = data['y']

        difX = []
        difY = []
        tdif = []


        for i in range(len(data) - 1):
            difX.append(float(Xs[i+1]) - float(Xs[i]))
            difY.append(float(Ys[i+1]) - float(Ys[i]))
            tdif.append(float(l_time[i+1]) - float(l_time[i]))

        dif = np.sqrt(np.power(difX,2) + np.power(difY,2)) #in pix
        #print (dif)

        velocity = dif / tdif
        #print velocity in pix/sec
        #print tdif

        mvmts = [] #length is len(data)-1

        for v in velocity:
            if (v < v_threshold):
                #fixation
                mvmts.append(1)
                #print v, v_threshold
            else:
                mvmts.append(0)

        fixations = []
        fs = []


        for m in range(len(mvmts)):
            if(mvmts[m] == 0):
                if(len(fs) > 0):
                    fixations.append(fs)
                    fs = []
            else:
                fs.append(m)

        if(len(fs) > 0):
            fixations.append(fs)

        #print fixations
        centroidsX = []
        centroidsY = []
        time0 = []
        time1 = []

        for f in fixations:
            cX = 0
            cY = 0

            if(len(f) == 1):
                i = f[0]
                cX = (float(data['x'][i]) + float(data['x'][i+1]))/2.0
                cY = (float(data['y'][i]) + float(data['y'][i+1]))/2.0
                t0 = float(data['timestamp'][i])
                t1 = float(data['timestamp'][i+1])

            else:
                t0 = float(data['timestamp'][f[0]])
                t1 = float(data['timestamp'][f[len(f)-1]+1])

                for e in range(len(f)):

                    cX += float(data['x'][f[e]]) 
                    cY += float(data['y'][f[e]])

                cX += float(data['x'][f[len(f)-1]+1]) 
                cY += float(data['y'][f[len(f)-1]+1]) 

                cX = cX / float(len(f)+1)
                cY = cY / float(len(f)+1)

            centroidsX.append(cX)
            centroidsY.append(cY)
            time0.append(t0)
            time1.append(t1)

        #create dataframe
        cxy_df = pd.DataFrame(np.column_stack([centroidsX, centroidsY, time0, time1]), 
                     columns=['cx', 'cy', 'start','end'])

        return cxy_df

    def hmm(self, data, filter_type, config):
        """
        Hidden Makov Model, adapted from https://gitlab.com/nslr/nslr-hmm.

        Parameters
        ----------		
        data : :class:`pandas.DataFrame`
            Pandas dataframe of x,y and timestamp positions.
        filter_type : :class:`dict`
            Types of filters to use.
		config : :class:`dict`
            Configuration data for data analysis. i.e. trial number, location.

        Attributes
        ----------
        data : :class:`numpy.ndarray`
            The smoothed signal (or it's n-th derivative).
        dr_th : :obj:`str`
			Data threshold.

        Notes
		-----
        Saccade:  
			The saccade is a ballistic movement, meaning it is pre-programmed and
        	does not change once it has started. Saccades of amplitude 40° peak at velocities
        	of 300–600°/s and last for 80–150 ms.
        Fixation:  
			the point between any two saccades, during which the eyes are relatively
			stationary and virtually all visual input occurs. Regular eye movement alternates
			between saccades and visual fixations, the notable exception being in smooth pursuit.
		Smooth pursuit:  
			Smooth pursuit movements are much slower tracking movements of
			the eyes designed to keep a moving stimulus on the fovea. Such movements are under
			voluntary control in the sense that the observer can choose whether or not to
			track a moving stimulus. (Neuroscience 2nd edition).

		References
        ----------
        .. [1] Pekkanen, J., & Lappi, O. (2017). A new and general approach to signal denoising and eye movement
			classification based on segmented linear regression. Scientific Reports, 7(1). doi:10.1038/s41598-017-17983-x.
        """

        from . import nslr_hmm

        t = data['timestamp'].values
        x = data['x']
        y = data['y']

        #class_df (list of fixations), c_sample (list of samples with associated classification)
        class_df, c_sample = nslr_hmm.classify_gaze(t, np.vstack((x, y)).T)

        #keep fixation events only
        class_df = class_df[(class_df['class'] == 1)]

        #if trial has at least one fixation
        if class_df.shape[0]>=1:
            ###split cxy0 tuple into cx0 and cy0
            class_df[['cx0', 'cy0']] = class_df['cxy0'].apply(pd.Series) 
            ###split cxy0 tuple into cx1 and cy1 
            class_df[['cx1', 'cy1']] = class_df['cxy1'].apply(pd.Series)

            ###drop column
            class_df = class_df.drop(['cxy0','cxy1'], axis=1)

            ###merge c_sample
            data['%s_class'%(filter_type)] = c_sample

        return data, class_df

    def idt(self,data, dis_threshold, dur_threshold):
        """
        Identification with Dispersion Threshold.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            The smoothed signal (or it's n-th derivative).
        dr_th : :obj:`str`
            Fixation duration threshold in pix/msec
        di_th : :obj:`str`
            Dispersion threshold in pixels

        Returns
        -------
        ys : :class:`numpy.ndarray`
            The smoothed signal (or it's n-th derivative).

        Notes
        -----
        The I-DT algorithm has two parameters: a dispersion threshold and
        the length of a time window in which the dispersion is calculated. The length of the
        time window is often set to the minimum duration of a fixation, which is around
        100-200 ms.

        """

        current = 0 #pointer to represent the current beginning point of the window
        last = 0
        #final lists for fixation info
        centroidsX = []
        centroidsY = []
        time0 = []
        time1 = []

        while (current < len(data)):

            t0 = float(data['timestamp'][current]) #beginning time
            t1 = t0 + float(dur_threshold)     #time after a min. fix. threshold has been observed

            for r in range(current, len(data)): 
                if(float(data['timestamp'][r])>= t0 and float(data['timestamp'][r])<= t1):
                    #print "if",r
                    last = r #this will find the last index still in the duration threshold


            #now check the dispersion in this window
            #print "window", current, last
            points = data[current:last+1]
            dispersion = 0

            argxmin = np.min(points.loc[:,'x'].astype(np.float))
            argxmax = np.max(points.loc[:,'x'].astype(np.float))

            argymin = np.min(points.loc[:,'y'].astype(np.float))
            argymax = np.max(points.loc[:,'y'].astype(np.float))

            dispersion = ((argxmax - argxmin) + (argymax - argymin))/2

            if (dispersion <= dis_threshold):

                #add new points
                while(dispersion <= dis_threshold and last + 1 < len(data)):

                    last += 1

                    points = data[current:last+1]
                    dispersion = 0

                    argxmin = np.min(points.loc[:,'x'].astype(np.float))
                    argxmax = np.max(points.loc[:,'x'].astype(np.float))

                    argymin = np.min(points.loc[:,'y'].astype(np.float))
                    argymax = np.max(points.loc[:,'y'].astype(np.float))

                    dispersion = ((argxmax - argxmin) + (argymax - argymin))/2

                #dispersion threshold is exceeded
                #fixation at the centroid [current,last]

                cX = 0
                cY = 0

                for f in range(current, last + 1):
                    cX += float(data['x'][f])
                    cY += float(data['y'][f])

                cX = cX / float(last - current + 1)
                cY = cY / float(last - current + 1)

                t0 = float(data['timestamp'][current])
                t1 = float(data['timestamp'][last])

                centroidsX.append(cX)
                centroidsY.append(cY)
                time0.append(t0)
                time1.append(t1)

                current = last + 1 #this will move the pointer to a novel window

            else:
                current += 1 #this will remove the first point
                last = current #this is not necessary

        #create dataframe
        cxy_df = pd.DataFrame(np.column_stack([centroidsX, centroidsY, time0, time1]), 
                     columns=['cx', 'cy', 'start','end'])

        return cxy_df

    def simple(self,df, missing, maxdist, mindur):
        """Detects fixations, defined as consecutive samples with an inter-sample
        distance of less than a set amount of pixels (disregarding missing data).
   
        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of x,y and timestamp positions
        missing : :obj:`str`
            Value to be used for missing data (default = 0.0)
        maxdist : :obj:`str`
            Maximal inter-sample distance in pixels (default = 25)
        mindur : :obj:`str`
            Minimal duration of a fixation in milliseconds; detected fixation cadidates will be 
            disregarded if they are below this duration (default = 100).

        Returns
        -------
        Sfix : :class:`numpy.ndarray` shape (N)
            list of lists, each containing [starttime]
        Efix : :class:`numpy.ndarray` shape (N)
            list of lists, each containing [starttime, endtime, duration, endx, endy]

        Notes
        -----
        From https://github.com/esdalmaijer/PyGazeAnalyser/blob/master/pygazeanalyser/detectors.py

        """
        x = df['x']
        y = df['y']
        time = df['timestamp']
        # empty list to contain data
        Sfix = []
        Efix = []

        # loop through all coordinates
        si = 0
        fixstart = False
        for i in range(1,len(x)):
            # calculate Euclidean distance from the current fixation coordinate
            # to the next coordinate
            dist = ((x[si]-x[i])**2 + (y[si]-y[i])**2)**0.5
            # check if the next coordinate is below maximal distance
            if dist <= maxdist and not fixstart:
                # start a new fixation
                si = 0 + i
                fixstart = True
                Sfix.append([time[i]])
            elif dist > maxdist and fixstart:
                # end the current fixation
                fixstart = False
                # only store the fixation if the duration is ok
                if time[i-1]-Sfix[-1][0] >= mindur:
                    Efix.append([Sfix[-1][0], time[i-1], time[i-1]-Sfix[-1][0], x[si], y[si]])
                # delete the last fixation start if it was too short
                else:
                    Sfix.pop(-1)
                si = 0 + i
            elif not fixstart:
                si += 1


        #return Sfix, Efix
        return Sfix, Efix




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Interface for the SR Research Eyelink eyetracking system.  
| `@date`: Created on Wed Feb 13 15:37:43 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['Eyelink']

# required external libraries
__required__ = ['psychopy','platform','pandas','pathlib']

# core
import time
import os
import re
import platform
import pandas as pd
from pathlib import Path

# debug
from pdb import set_trace as breakpoint

# local libraries
if __name__ == '__main__':
	from .. import settings
	from . import Calibration
	from . import pylink

class Eyelink():
    """Interface for the SR Research Eyelink eyetracking system."""
    def __init__(self, window, timer, isPsychopy=True, subject=None, **kwargs):
        """
        Initiate the mdl.eyetracking.Eyelink module.

        Parameters
        ----------
        window :  `psychopy.visual.Window <https://www.psychopy.org/api/visual/window.html#window>`_
            PsychoPy window instance.
        timer :  `psychopy.core.CountdownTimer <https://www.psychopy.org/api/core.html#psychopy.core.CountdownTimer>`_
            Psychopy timer instance.
        isPsychopy : :obj:`bool`
            Is Psychopy being used. Default `True`.
        subject : :obj:`int`
            Subject number.
        **kwargs : :obj:`str` or :obj:`None`, optional
            Here's a list of available properties:

            .. list-table::
                :class: kwargs
                :widths: 25 50
                :header-rows: 1

                * - Property
                  - Description
                * - **isFlag** : :obj:`bool`
                  - Bypass Eyelink flags (`isRecording`, `isConnected`) to run all functions without checking flags. Default is `True`.
                * - **isLibrary** : :obj:`bool`
                  - Check if required packages have been installed. Default is `False`.
                * - **demo** : :obj:`bool`
                  - Run demo mode, which includes region of interest highlighting and other testing methods. Default is `False`.

        Examples
        --------
        >>> eytracking = mdl.eyetracking(window=window, subject=subject)

	    Notes
	    -----
	    According to `pylink.chm <../manual/pylink.chm>`_, the sequence of operations for implementing a trial is:
	        1) Perform a DRIFT CORRECTION, which also serves as the pre-trial fixation target.
	        2) Start recording, allowing 100 milliseconds of data to accumulate before the trial display starts.
	        3) Draw the subject display, recording the time that the display appeared by placing a message in the EDF file.
	        4) Loop until one of these events occurs RECORDING halts, due to the tracker ABORT menu or an error, the maximum \
	        trial duration expires 'ESCAPE' is pressed, the program is interrupted, or abutton on the EyeLink button box is pressed.
	        5) Add special code to handle gaze-contingent display updates.
	        6) Blank the display, stop recording after an additional 100 milliseconds of data has been collected.
	        7) Report the trial result, and return an appropriate error code.
        """
        # psychopy
        from psychopy import visual, core, event
        from psychopy.constants import (NOT_STARTED, STARTED, FINISHED)

        #----introduction
        settings.console(msg="mdl.eyetracking() found.", c='green')

        #----screen size
        if isPsychopy:
            self.w = int(window.size[0])
            self.h = int(window.size[1])
        else:
            # if windows
            if platform.system() == "Windows":
				## try to import
                try:
                    from win32api import GetSystemMetrics
                except ImportError:
                    self.console("No module named 'win32api'. Installing from PyPI",'red')
                    settings.library(['win32api'])
				# get width, height	
                self.w = int(GetSystemMetrics(0))
                self.h = int(GetSystemMetrics(1))
                settings.console(msg="mdl.eyetracking(): screensize [%s, %s]."%(self.w, self.h))
            # else if osx
            elif platform.system() == 'Darwin':
                ## try to import
                try:
                    import pyobjc
                    from AppKit import NSScreen
                except ImportError:
                    self.console("No module named 'pyobjc'. Installing from PyPI",'red')
                    settings.library(['pyobjc'])
				# get width, height	
                self.w = int(NSScreen.mainScreen().frame().size.width)
                self.h = int(NSScreen.mainScreen().frame().size.height)
                settings.console(msg="mdl.eyetracking(): screensize [%s, %s]."%(self.w, self.h))

        #----parameters
        # instants
        self.window = window
        # flags
        self.isConnected = False
        self.isStarted = False
        self.isCalibration = False
        self.isRecording = False
        self.isDriftCorrection = False
        self.isFinished = False
        # counters
        self.drift_count = 0
        # metadata
        self.trial = ''
        self.block = ''
        # pupil
        self.eye_used = None
        self.left_eye = 0
        self.right_eye = 1
        # status
        self.status = {}

        #----timing
        self.routineTimer = timer

        #----drift correction
        #clock
        self.drift_message_clock = core.Clock()
        self.drift_clock = core.Clock()
        #screen
        text = '\n'.join([
            "On the next screen a fixation dot will appear.",
            "Please look at the fixation dot until it disappears.",
            "\nPress any key to continue."
        ])
        self.drift_text = visual.TextStim(win=window, name='drift_message', font='Helvetica',
        text=text, height=0.1, wrapWidth=1.5, ori=0, pos=(0, 0), alignVert='center', color='black', 
        colorSpace='rgb', opacity=1, depth=0.0)
        
        # check if subject number has been entered
        if subject == None:
            raise AssertionError('Subject number not entered. Please enter the subject number.')
        else:
            self.subject = subject

        #----edf filename
        self.subject = os.path.splitext(str(subject))[0]
        # check if integer
        if re.match(r'\w+$', self.subject):
            pass
        else:
            self.window.close()
            raise AssertionError('Name must only include A-Z, 0-9, or _')
        # check length
        if len(self.subject) <= 8:
            pass
        else:
            self.window.close()
            raise AssertionError('Name must be <= 8 characters.')
        # store name
        self.fname = str(self.subject) + '.edf'

        # generate file path
        self.path = "%s/data/edf/" % (os.getcwd())
        if not os.path.exists(self.path):
            os.makedirs(self.path)

		#----kwargs
        # demo
        self.isDemo = kwargs['isDemo'] if 'isDemo' in kwargs else False
        # turn off eyelink flags
        self.isFlag = kwargs['isFlag'] if 'isFlag' in kwargs else True
        # isLibrary
        self.isLibrary = kwargs['isLibrary'] if 'isLibrary' in kwargs else False

        #----check if required libraries are available
        if self.isLibrary:
            settings.library(__required__)

        #----kwargs

    def connect(self, calibration_type=13, automatic_calibration_pacing=1000, saccade_velocity_threshold=35,
                saccade_acceleration_threshold=9500, sound=True, select_parser_configuration=0,
                recording_parse_type="GAZE", enable_search_limits=True, ip='100.1.1.1'):
        """
        Connect to Eyelink.

        Parameters
        ----------
        ip : :obj:`string`
            Host PC ip address.
        calibration_type : :obj:`int`
            Calibration type. Default is 13-point. [see `Eyelink 1000 Plus User Manual, 3.7 Calibration \
			<../manual/EyeLink%201000%20Plus%20User%20Manual%201.0.12.pdf>`_]
        automatic_calibration_pacing : :obj:`int`
            Select the delay in milliseconds, between successive calibration or validation targets 
            if automatic target detection is activeSet automatic calibration pacing. [see `pylink.chm <../manual/pylink.chm>`_]
        saccade_velocity_threshold : :obj:`int`
            Sets velocity threshold of saccade detector: usually 30 for cognitive research, 22 for 
            pursuit and neurological work. Default is 35. Note: EyeLink II and EyeLink 1000,
            select select_parser_configuration should be used instead. [see `EyeLink Programmer’s Guide, \
            Section 25.9: Parser Configuration <../manual/EyeLink%20Programmers%20Guide.pdf>`_; `Eyelink 1000 Plus \
			User Manual, Section 4.3.5 Saccadic Thresholds <../manual/EyeLink%201000%20Plus%20User%20Manual%201.0.12.pdf>`_]
        saccade_acceleration_threshold : :obj:`int`
            Sets acceleration threshold of saccade detector: usually 9500 for cognitive research, 5000 
            for pursuit and neurological work. Default is 9500. Note: For EyeLink II and EyeLink 1000,
            select select_parser_configuration should be used instead. [see `EyeLink Programmer’s Guide, \
            Section 25.9: Parser Configuration <../manual/EyeLink%20Programmers%20Guide.pdf>`_; \
			`Eyelink 1000 Plus User Manual, Section 4.3.5 Saccadic Thresholds <../manual/EyeLink%201000%20Plus%20User%20Manual%201.0.12.pdf>`_]
        select_parser_configuration : :obj:`int`
            Selects the preset standard parser setup (0) or more sensitive (1). These are equivalent
            to the cognitive and psychophysical configurations. Default is 0. [see `EyeLink Programmer’s \
            Guide, Section 25.9: Parser Configuration <../manual/EyeLink%20Programmers%20Guide.pdf>`_]
        sound : :obj:`bool`
            Should sound be used for calibration/validation/drift correction.
        recording_parse_type : :obj:`str`
            Sets how velocity information for saccade detection is to be computed.
            Enter either 'GAZE' or 'HREF'. Default is 'GAZE'. [see `Eyelink 1000 Plus User Manual, \
			Section 4.4: File Data Types <../manual/EyeLink%201000%20Plus%20User%20Manual%201.0.12.pdf>`_]
        enable_search_limits : :obj:`bool`
            Enables tracking of pupil to global search limits. Default is True. [see `Eyelink 1000 Plus User Manual, \
			Section 4.4: File Data Types <../manual/EyeLink%201000%20Plus%20User%20Manual%201.0.12.pdf>`_]
        
        Returns
        ----------
        param : :class:`pandas.DataFrame`
            Returns dataframe of parameters for subject.
        
        Examples
        --------
        >>> param = eyetracking.connect(calibration_type=13)
        """
        #----initiate connection with eyetracker
        self.ip = ip
        try:
            self.tracker = pylink.EyeLink(self.ip)
            self.isConnected = True
            settings.console(c='blue', msg="Eyelink Connected")
        except RuntimeError:
            self.tracker = pylink.EyeLink(None)
            self.isConnected = False
            settings.console(c='red', msg="Eyelink not detected at %s"%(self.ip))
            
        #----tracker metadata
        # get eyelink version
        self.tracker_version = self.tracker.getTrackerVersion()

        # get host tracking software version
        self.host_version = 0
        if self.tracker_version == 3:
            tvstr = self.tracker.getTrackerVersionString()
            vindex = tvstr.find("EYELINK CL")
            self.host_version = int(
                float(tvstr[(vindex + len("EYELINK CL")):].strip()))

        #----preset
        self.select_parser_configuration = select_parser_configuration
        self.saccade_acceleration_threshold = saccade_acceleration_threshold
        self.saccade_velocity_threshold = saccade_velocity_threshold
        self.recording_parse_type = recording_parse_type
        self.enable_search_limits = "YES" if enable_search_limits else "NO"
        # calibration
        self.calibration_type = calibration_type  # []-point calibration
        self.automatic_calibration_pacing = automatic_calibration_pacing
        # specify the EVENT and SAMPLE data that are stored in EDF or retrievable from the Link
        # [see Section 4.6 Data Files of the  EyeLink 1000 Plus user manual]
        self.fef = "LEFT,RIGHT,FIXATION,SACCADE,BLINK,MESSAGE,BUTTON,INPUT"
        self.lef = "LEFT,RIGHT,FIXATION,FIXUPDATE,SACCADE,BLINK,BUTTON,INPUT"
        if self.host_version >= 4:
            self.fsd = "LEFT,RIGHT,GAZE,AREA,GAZERES,STATUS,HTARGET,INPUT"
            self.lsd = "LEFT,RIGHT,GAZE,GAZERES,AREA,STATUS,HTARGET,INPUT"
        else:
            self.fsd = "LEFT,RIGHT,GAZE,AREA,GAZERES,STATUS"
            self.lsd = "LEFT,RIGHT,GAZE,GAZERES,AREA,STATUS"

        #---store param for export
        self.param = dict(
            tracker_version=self.tracker_version, 
            host_version=self.host_version,
            select_parser_configuration=self.select_parser_configuration,
            saccade_acceleration_threshold=self.saccade_acceleration_threshold,
            saccade_velocity_threshold=self.saccade_velocity_threshold,
            recording_parse_type=self.recording_parse_type,
            enable_search_limits=self.enable_search_limits,
            automatic_calibration_pacing=self.automatic_calibration_pacing,
            file_event_filter=self.fef,
            file_sample_data=self.fsd,
            link_event_filter=self.lef,
            link_sample_data=self.lsd,
            calibration_type=self.calibration_type
        )

        #----open edf
        self.tracker.openDataFile(self.fname)
        pylink.flushGetkeyQueue()

        #----send settings to eyelink
        # place EyeLink tracker in offline (idle) mode before changing settings
        self.tracker.setOfflineMode()

        # Set the tracker to parse events using either GAZE or HREF
        # note: [see `Data Viewer User Manual, Section 7: Protocol for EyeLink Data to Viewer Integration <../manual/EyeLink%20Data%20Viewer%20User%20Manual%203.2.1.pdf>`_]
        self.tracker.sendCommand("recording_parse_type = %s" %(self.recording_parse_type))
   
        # inform the tracker the resolution of the subject display
        # note: [see Eyelink Installation Guide, Section 8.4: Customizing Your PHYSICAL.INI Settings]
        self.tracker.sendCommand("screen_pixel_coords = 0 0 %d %d" % (self.w - 1, self.h - 1))
   
        # save display resolution in EDF data file for Data Viewer integration purposes
        # note: [see `Data Viewer User Manual, Section 7: Protocol for EyeLink Data to Viewer Integration <../manual/EyeLink%20Data%20Viewer%20User%20Manual%203.2.1.pdf>`_]
        self.tracker.sendMessage("DISPLAY_COORDS = 0 0 %d %d" % (self.w - 1, self.h - 1))

        # set calibration type
        self.tracker.sendCommand("calibration_type = HV%d"%(self.calibration_type))
        # set automatic calibraiton pacing interval
        self.tracker.sendCommand("automatic_calibration_pacing = %d" % (
            self.automatic_calibration_pacing))

        # if using tracker version 2 or greater
        if self.tracker_version >= 2:
            self.tracker.sendCommand("select_parser_configuration %d" % (
                self.select_parser_configuration))
            # turn of scene link
            if self.tracker_version == 2:
                self.tracker.sendCommand("scene_camera_gazemap = NO")
        else:
            self.tracker.sendCommand("saccade_velocity_threshold = %s" % (
                self.saccade_velocity_threshold))
            self.tracker.sendCommand("saccade_acceleration_threshold = %s" % (
                self.saccade_acceleration_threshold))

        # if using tracker version 3 or above
        if self.tracker_version >= 3:
            self.tracker.sendCommand(
                "enable_search_limits=%s" % (self.enable_search_limits))
            self.tracker.sendCommand("track_search_limits=YES")
            self.tracker.sendCommand("autothreshold_click=YES")
            self.tracker.sendCommand("autothreshold_repeat=YES")
            self.tracker.sendCommand("enable_camera_position_detect=YES")

        # set content of edf file
        # edf filters #event types
        self.tracker.sendCommand('file_event_filter = %s' % (self.fef))
        self.tracker.sendCommand('file_sample_data = %s' % (self.fsd))
        self.tracker.sendCommand('link_event_filter = %s' % (self.lef))
        self.tracker.sendCommand('link_sample_data = %s' % (self.lsd))

        # program button '5' for drift correction
        self.tracker.sendCommand("button_function 5 'accept_target_fixation'")

        # select sound for calibration and drift correct
        # "" = sound, "off" = no sound
        pylink.setCalibrationSounds("", "", "")
        pylink.setDriftCorrectSounds("", "", "")
        
        # export param to txt file
        ## prevent metadata from being truncated
        width = pd.get_option('display.max_colwidth')
        pd.set_option('display.max_colwidth', -1)
        # create df
        param = pd.DataFrame(list(self.param.items()),columns=['category', 'value'])
        param.to_csv(path_or_buf=self.path + str(self.subject) + ".txt", index=True, index_label='index')
        # reset truncation
        pd.set_option('display.max_colwidth', width)

        #if ipython		
		# from IPython.display import display
        # display(param)
        
        return param

    def set_eye_used(self, eye):
        """
        Set dominant eye. This step is required for recieving gaze coordinates from Eyelink->Psychopy.

        Parameters
        ----------
        eye : :obj:`str`
            Dominant eye (left, right). This will be used for outputting Eyelink gaze samples.
        
        Examples
        --------
        >>> dominant_eye = 'left'
        >>> eye_used = eyetracking.set_eye_used(eye=dominant_eye)
        """
        settings.console(msg="eyetracking.set_eye_used()")
        eye_entered = str(eye)
        if eye_entered in ('Left', 'LEFT', 'left', 'l', 'L'):
            settings.console(c="blue", msg="eye_entered = %s('left')" %(eye_entered))
            self.eye_used = self.left_eye
        else:
            settings.console(c="blue", msg="eye_entered = %s('right')" %(eye_entered))
            self.eye_used = self.right_eye

        return self.eye_used

    def calibration(self):
        """
        Start calibration procedure.
        
        Returns
        -------
        isCalibration : :obj:`bool`
            Message indicating status of calibration.
        
        Examples
        --------
        >>> eyetracking.calibration()
        """		
        # calibration
        from . import Calibration

        settings.console(msg="eyetracking.calibration()")
        
        #----if connected to eyetracker
        if self.isConnected or (not self.isFlag):
            # put the tracker in offline mode before we change its configurations
            self.tracker.setOfflineMode()
            # Generate custom calibration stimuli
            self.genv = Calibration(w=self.w, h=self.h, tracker=self.tracker, window=self.window)
            # execute custom calibration display
            pylink.openGraphicsEx(self.genv)
            # calibrate
            self.tracker.doTrackerSetup(self.w, self.h)
            #flip screen after finishing
            self.window.flip()
            
        #----finished
        self.isCalibration = True
        settings.console(c="blue", msg="eyetracking.calibration() finished")

        return self.isCalibration
    
    def _drift_message(self):
        #----prepare to start Routine 'drift_message'
        endExpNow = False  # flag for 'escape' or other condition => quit the exp
        t = 0
        self.drift_message_clock.reset()  # clock
        frameN = -1
        continueRoutine = True
        # update component parameters for each repeat
        drift_response = event.BuilderKeyResponse()
        # keep track of which components have finished
        drift_message_components = [drift_response, self.drift_text]
        for thisComponent in drift_message_components:
            if hasattr(thisComponent, 'status'):
                thisComponent.status = NOT_STARTED
        
        # -------Start Routine "drift_message"-------
        while continueRoutine:
            # get current time
            t = self.drift_message_clock.getTime()
            frameN = frameN + 1  # number of completed frames (so 0 is the first frame)
            # update/draw components on each frame
            
            # *drift_response* updates
            if t >= 0.0 and drift_response.status == NOT_STARTED:
                # keep track of start time/frame for later
                drift_response.tStart = t
                drift_response.frameNStart = frameN  # exact frame index
                drift_response.status = STARTED
                # keyboard checking is just starting
                event.clearEvents(eventType='keyboard')
            if drift_response.status == STARTED:
                theseKeys = event.getKeys()
                
                # check for quit:
                if "escape" in theseKeys:
                    endExpNow = True
                if len(theseKeys) > 0:  # at least one key was pressed
                    # a response ends the routine
                    continueRoutine = False
            
            # *drift_text* updates
            if t >= 0.0 and self.drift_text.status == NOT_STARTED:
                # keep track of start time/frame for later
                self.drift_text.tStart = t
                self.drift_text.frameNStart = frameN  # exact frame index
                self.drift_text.setAutoDraw(True)
            
            # check for quit (typically the Esc key)
            if endExpNow or event.getKeys(keyList=["escape"]):
                core.quit()
            
            # check if all components have finished
            if not continueRoutine:  # a component has requested a forced-end of Routine
                break
            continueRoutine = False  # will revert to True if at least one component still running
            for thisComponent in drift_message_components:
                if hasattr(thisComponent, "status") and thisComponent.status != FINISHED:
                    continueRoutine = True
                    break  # at least one component has not yet finished
            
            # refresh the screen
            if continueRoutine:  # don't flip if this routine is over or we'll get a blank screen
                self.window.flip()
        
        # -------Ending Routine "drift_message"-------
        for thisComponent in drift_message_components:
            if hasattr(thisComponent, "setAutoDraw"):
                thisComponent.setAutoDraw(False)
                
        # the Routine "drift_message" was not non-slip safe, so reset the non-slip timer
        self.routineTimer.reset()
        
    def drift_correction(self, origin='call'):
        """
        Starts drift correction. This can be done at any point after calibration, including before 
        and after eyetracking.start_recording has already been initiated.
        
        Parameters
        ----------
        origin : :obj:`str`
            Origin of call, either `manual` (default) or from gaze contigent event (`gc`).
        
        Returns
        -------
        isDriftCorrection : :obj:`bool`
            Message indicating status of drift correction.
        
        Notes
        -----
        Running drift_correction will end any start_recording event to function properly. Once drift
        correction has occured, it is safe to run start_recording.
        
        Examples
        --------
        >>> eyetracking.drift_correction()
        """
        settings.console(msg="eyetracking.drift_correction()")
        
        #----present drift correction message (only if manually accessing)
        if origin=='call':
            self._drift_message()
            self.window.clearBuffer()
            self.window.flip()
        
        #check if recording
        if self.isRecording or (not self.isFlag):
            # end of trial message
            self.tracker.sendMessage('drift correction')

            # end realtime mode
            pylink.endRealTimeMode()
            pylink.msecDelay(100)

            # send trial-level variables
            #if running from self.gc
            if origin=='gc':
                self.send_variable(variables=dict(trial=self.trial, block=self.block, issues='gc window failed'))

            # specify end of trial
            self.tracker.sendMessage("TRIAL_RESULT 1")

            # stop recording eye data
            self.tracker.stopRecording()
        
        # flag isRecording
        self.isRecording = False
        
        # initiate drift correction, flag isDriftCorrection
        self.isDriftCorrection = True
        try:
           self.tracker.doDriftCorrect(int(self.w/2), int(self.h/2), 1, 1)
        except:
           self.tracker.doTrackerSetup()

        #----finished
        settings.console(c="blue", msg="eyetracking.drift_correction() finished")
        
        return self.isDriftCorrection
    
    def start_recording(self, trial, block):
        """
        Starts recording of Eyelink.

        Parameters
        ----------
        trial : :obj:`str`
            Trial Number.
        block : :obj:`str`
            Block Number.
        
        Returns
        -------
        isRecording : :obj:`bool`
            Message indicating status of Eyelink recording.

        Notes
        -----
        tracker.beginRealTimeMode():
            To ensure that no data is missed before the important part of the trial starts. The EyeLink 
            tracker requires 10 to 30 milliseconds after the recording command to begin writing data. 
            This extra data also allows the detection of blinks or saccades just before the trial start, 
            allowing bad trials to be discarded in saccadic RT analysis. A "SYNCTIME" message later 
            in the trial marks the actual zero-time in the trial's data record for analysis.
            [see `pylink.chm <../manual/pylink.chm>`_]
        TrialID:
            The "TRIALID" message is sent to the EDF file next. This message must be placed 
            in the EDF file before the drift correction and before recording begins, and is critical 
            for data analysis. The viewer will not parse any messages, events, or samples that exist
            in the data file prior to this message. The command identifier can be changed in the data
            loading preference settings.
            [see `Data Viewer User Manual, Section 7: Protocol for EyeLink Data to Viewer Integration \
			<../manual/EyeLink%20Data%20Viewer%20User%20Manual%203.2.1.pdf>`_]
        SYNCTIME:
            Marks the zero-time in a trial. A number may follow, which is interpreted as the delay of 
            the message from the actual stimulus onset. It is suggested that recording start 100 
            milliseconds before the display is drawn or unblanked at zero-time, so that no data at the 
            trial start is lost.
            [see `pylink.chm <../manual/pylink.chm>`_]

        Examples
        --------
        >>> eyetracking.start_recording(trial=1, block=1)
        """
        settings.console(msg="eyetracking.start_recording()")
        
        # indicates start of trial
        self.tracker.sendMessage('TRIALID %s' % (str(trial)))
    
        # Message to post to Eyelink display Monitor
        self.tracker.sendCommand("record_status_message 'Trial %s Block %s'" % (str(trial), str(block)))

        # start recording, parameters specify whether events and samples are
        # stored in file, and available over the link
        self.tracker.startRecording(1, 1, 1, 1)

        # buffer to prevent loss of data
        pylink.beginRealTimeMode(100)

        # indicates zero-time of trial
        self.tracker.sendMessage('SYNCTIME')
        self.tracker.sendMessage('start recording')
            
        # flag isDriftCorrection, isRecording, trial, block
        self.isDriftCorrection = False
        self.isRecording = True
        self.trial = trial
        self.block = block

        return self.isRecording
    
    def gc(self, bound, min_, max_=None):
        """
        Creates gaze contigent event. This function needs to be run while recording.
        
        Parameters
        ----------
        bound : :obj:`dict` [:obj:`str`, :obj:`int`]:
            Dictionary of the bounding box for each region of interest. Keys are each side of the 
            bounding box and values are their corresponding coordinates in pixels.
        min_ : :obj:`int`
            Mininum duration (msec) in which gaze contigent capture is collecting data before allowing
            the task to continue.
        max_ : :obj:`int` or :obj:`None`
            Maxinum duration (msec) before task is forced to go into drift correction. 

        Examples
        --------
        >>> # Collect samples within the center of the screen, for 2000 msec, 
        >>> # with a max limit of 10000 msec.
        >>> region = dict(left=860, top=440, right=1060, bottom=640)
        >>> eyetracking.gc(bound=bound, min_=2000, max_=10000)
        """
        
        #if eyetracker is recording
        if self.isRecording or (not self.isFlag):
            # if demo
            if self.isDemo:
                line = 6
                width = (bound['right'] - bound['left']) + line
                height = (bound['bottom'] - bound['top']) + line
                center = ((bound['left'] - line) + width/2, bound['top'] + height/2)
                color = [0,255,0]
                box = visual.Rect(win=self.window, name="bound", units='pix',
                                  width=width, height=height, pos=(0,0), colorSpace='rgb255',
                                  lineWidth=6, lineColor=color, lineColorSpace='rgb255',
                                  opacity=1, depth=0.0, interpolate=True)
                box.setAutoDraw(True)
                
            # draw box
            self.tracker.sendCommand('draw_box %d %d %d %d 6'% (bound['left'],bound['top'],bound['right'],bound['bottom']))
            
            #get start time
            start_time = time.clock()
            #get current time
            current_time = time.clock()
            
            #----check gc window
            while True:    
                #flip screen
                self.window.flip()
                
                #get gaze sample
                gxy, ps, s = self.sample()
                
                #is gaze with gaze contigent window for min_ time
                if ((bound['left'] < gxy[0] < bound['right']) and (bound['top'] < gxy[1] < bound['bottom'])):
                    # if demo
                    if self.isDemo:
                        box.lineColor = [0,255,0]
                    # has mininum time for gc window occured
                    duration = (time.clock() - current_time) * 1000
                    if duration > min_:
                        settings.console(c='blue', msg="eyetracking.gc() success in %d msec"%((time.clock() - start_time) * 1000))
                        self.send_message(msg='gc window success')
                        # clear bounding box (if demo)
                        if self.isDemo:
                            box.setAutoDraw(False)
                        break
                # not in window
                else:
                    # if demo
                    if self.isDemo:
                        box.lineColor = [255,0,0]
                    #reset current time
                    current_time = time.clock()
                
                # if reached maxinum time
                if max_ is not None:
                    # has maxinum time for gc window occured
                    duration = (time.clock() - start_time) * 1000
                    if duration > max_:
                        settings.console(c='blue', msg="eyetracking.gc() failed, drift correction started")
                        self.send_message(msg='gc window failed')
                        # clear bounding box (if demo)
                        if self.isDemo:
                            box.setAutoDraw(False)
                        # start drift correction
                        self.drift_correction(origin='gc')
                        break
        else:
            settings.console(c='red', msg="eyetracker not recording")
                     
    def sample(self):
        """
        Collects new gaze coordinates from Eyelink.
        
        Returns
        -------
        gxy : :class:`tuple`
            Gaze coordiantes.
        ps : :obj:`tuple`
            Pupil size (area).
        sample : :class:`EyeLink.getNewestSample`
            Eyelink newest sample.

        Examples
        --------
        >>> eyetracking.sample(eye_used=eye_used)
        """
        # check for new sample update
        sample = self.tracker.getNewestSample()
        
        #if sample available
        if(sample != None):
            # if dominant eye entered as right in both psychopy and eyelink
            if ((self.eye_used == self.right_eye) and (sample.isRightSample())):
                gxy = sample.getRightEye().getGaze()
                ps = sample.getRightEye().getPupilSize()
            elif ((self.eye_used == self.left_eye) and (sample.isLeftSample())):
                gxy = sample.getLeftEye().getGaze()
                ps = sample.getLeftEye().getPupilSize()
            else:
                 raise AssertionError('eye_used (%s) ≠ Eyelink "Eye Tracked" (%s). \
                                      Please make sure the eye used in PsychoPy and Eyelink match.'\
                                      %(self.eye_used, sample.getEye()))
                
            # gaze coordinates, pupil area    
            return gxy, ps, sample
        
        return None, None, sample

    def send_message(self, msg):
        """
        Send message to Eyelink. This allows post-hoc processing of event markers (i.e. "stimulus onset").

        Parameters
        ----------
        msg : :obj:`str`
            Message to be recieved by eyelink.

        Examples
        --------
        >>> eyetracking.console(c="blue", msg="eyetracking.calibration() started")
        """
        self.tracker.sendMessage(msg)

    def send_variable(self, variables):
        """
        send trial variable to eyelink at the end of trial.

        Parameters
        ----------
        variable : :obj:`dict` or :obj:`None`
            Trial-related variables to be read by eyelink.          
        """
        if variables is not None:
            for key, value in variables.items():
                msg = "!V TRIAL_VAR %s %s" % (key, value)
                self.tracker.sendMessage(msg)
                pylink.msecDelay(1)
            # finished
            settings.console(msg="variables sent")
        else:
            settings.console(c="red", msg="no variables entered")

    def stop_recording(self, trial=None, block=None, variables=None):
        """
        Stops recording of Eyelink. Also allows transmission of trial-level variables to Eyelink.

        Parameters
        ----------
        trial : :obj:`int`
            Trial Number.
        block : :obj:`int`
            Block Number.
        variables : :obj:`dict` or :obj:`None`
            Dict of variables to send to eyelink (variable name, value).
        
        Returns
        -------
        isRecording : :obj:`bool`
            Message indicating status of Eyelink recording.

        Notes
        -----
        pylink.pumpDelay():
            Does a unblocked delay using currentTime(). This is the preferred delay function 
            when accurate timing is not needed.
            [see `pylink.chm <../../pylink.chm>`_]
        pylink.msecDelay():
            During calls to pylink.msecDelay(), Windows is not able to handle messages. One result of 
            this is that windows may not appear. This is the preferred delay function when accurate 
            timing is needed.
            [see `pylink.chm <../../pylink.chm>`_]
        tracker.endRealTimeMode():
            Returns the application to a priority slightly above normal, to end realtime mode. This 
            function should execute rapidly, but there is the possibility that Windows will allow other 
            tasks to run after this call, causing delays of 1-20 milliseconds. This function is equivalent 
            to the C API void end_realtime_mode(void).
            [see `pylink.chm <../../pylink.chm>`_]
        TRIAL_VAR: 
            Lets users specify a trial variable and value for the given trial. One
            message should be sent for each trial condition variable and its corresponding value. If
            this command is used there is no need to use TRIAL_VAR_LABELS. The default command identifier 
            can be changed in the data loading preference settings. Please note that the eye tracker 
            can handle about 20 messages every 10 milliseconds. So be careful not to send too many 
            messages too quickly if you have many trial condition messages to send. Add one millisecond 
            delay between message lines if this is the case.
            [see `pylink.chm <../../pylink.chm>`_]
        TRIAL_RESULT: 
            Defines the end of a trial. The viewer will not parse any messages, events, or samples that 
            exist in the data file after this message. The command identifier can be changed in the 
            data loading preference settings.
            [see `Data Viewer User Manual, Section 7: Protocol for EyeLink Data to Viewer Integration \
			<../manual/EyeLink%20Data%20Viewer%20User%20Manual%203.2.1.pdf>`_]

        Examples
        --------
        >>> variables = dict(stimulus='face.png', event='stimulus')
        >>> eyetracking.stop_recording(trial=trial, block=block, variables=variables)
        """
        settings.console(msg="eyetracking.stop_recording()")
        
        # end of trial message
        self.tracker.sendMessage('end recording')

        # end realtime mode
        pylink.endRealTimeMode()
        pylink.msecDelay(100)

        # send trial-level variables
        variables['trial'] = trial
        variables['block'] = block
        variables['issues'] = 'none'
        self.send_variable(variables=variables)

        # specify end of trial
        self.tracker.sendMessage("TRIAL_RESULT 1")

        # stop recording eye data
        self.tracker.stopRecording()

        # flag isRecording
        self.isRecording = False

        return self.isRecording

    def finish_recording(self, path=None):
        """
        Ending Eyelink recording.

        Parameters
        ----------
        path : :obj:`str` or :obj:`None`
            Path to save data. If None, path will be default from PsychoPy task.
        
        Returns
        -------
        isFinished : :obj:`bool`
            Message indicating status of Eyelink recording.

        Notes
        -----
        pylink.pumpDelay():
            Does a unblocked delay using currentTime(). This is the preferred delay function 
            when accurate timing is not needed.
            [see `pylink.chm <../manual/pylink.chm>`_]
        pylink.msecDelay():
            During calls to pylink.msecDelay(), Windows is not able to handle messages. One result of 
            this is that windows may not appear. This is the preferred delay function when accurate 
            timing is needed.
            [see `pylink.chm <../manual/pylink.chm>`_]
        tracker.setOfflineMode():
            Places EyeLink tracker in offline (idle) mode. Wait till the tracker has finished the 
            mode transition.
            [see `pylink.chm <../manual/pylink.chm>`_]
        tracker.endRealTimeMode():
            Returns the application to a priority slightly above normal, to end realtime mode. This 
            function should execute rapidly, but there is the possibility that Windows will allow other 
            tasks to run after this call, causing delays of 1-20 milliseconds. This function is 
            equivalent to the C API void end_realtime_mode(void).
            [see `pylink.chm <../manual/pylink.chm>`_]
        tracker.receiveDataFile():
            This receives a data file from the EyeLink tracker PC. Source filename and destination 
            filename should be given.
            [see `pylink.chm <../manual/pylink.chm>`_]

        Examples
        --------
        >>> #end recording session
        >>> eyetracking.finish_recording()
        """
        settings.console(msg="eyetracking.finish_recording()")
        
        #clear host display
        self.tracker.sendCommand('clear_screen 0') 
        
        # double check realtime mode has ended
        pylink.endRealTimeMode()
        pylink.pumpDelay(500)

        # rlaces eyeLink tracker in offline (idle) mode
        self.tracker.setOfflineMode()

        # allow buffer to prepare data for closing
        pylink.msecDelay(500)

        # closes any currently opened EDF file on the EyeLink tracker computer's hard disk
        self.tracker.closeDataFile()

        # This receives a data file from the eyelink tracking computer
        #if no path entered, use psychopy destination
        if path is None:
            destination = self.path + self.fname
        else:
            destination = path
        self.tracker.receiveDataFile(self.fname, destination)
        settings.console(c="blue", msg="File saved at: %s"%(Path(destination)))

        # sends a disconnect message to the EyeLink tracker
        self.tracker.close()

        # flag isFinished
        self.isFinished = True

        #---finish
        return self.isFinished
